# -*- coding: utf-8 -*-
"""
CODING GUIDELINES:

Do NOT (without talking to trask):
- add another high level method for sending or receiving messages (like recv_eventual_msg_without_reply)
- add a service to the list of services below unless you're SURE all nodes will need it!
- serialize anything with pickle
"""

# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from pydantic import BaseSettings
from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base

# syft absolute
import syft as sy

# relative
from ....grid import GridURL
from ....lib import lib_ast
from ....logger import debug
from ....logger import error
from ....logger import info
from ....logger import traceback_and_raise
from ....util import get_subclasses
from ...common.message import EventualSyftMessageWithoutReply
from ...common.message import ImmediateSyftMessageWithReply
from ...common.message import ImmediateSyftMessageWithoutReply
from ...common.message import SignedEventualSyftMessageWithoutReply
from ...common.message import SignedImmediateSyftMessageWithReply
from ...common.message import SignedImmediateSyftMessageWithoutReply
from ...common.message import SignedMessage
from ...common.message import SyftMessage
from ...common.uid import UID
from ...io.location import Location
from ...io.location import SpecificLocation
from ...io.route import Route
from ...io.route import SoloRoute
from ...io.virtual import create_virtual_connection
from ..abstract.node import AbstractNode
from .action.exception_action import ExceptionMessage
from .action.exception_action import UnknownPrivateException
from .client import Client
from .metadata import Metadata
from .node_manager.redis_store import RedisStore
from .node_manager.setup_manager import SetupManager
from .node_service.auth import AuthorizationException
from .node_service.child_node_lifecycle.child_node_lifecycle_service import (
    ChildNodeLifecycleService,
)
from .node_service.get_repr.get_repr_service import GetReprService
from .node_service.heritage_update.heritage_update_service import HeritageUpdateService
from .node_service.msg_forwarding.msg_forwarding_service import (
    SignedMessageWithReplyForwardingService,
)
from .node_service.msg_forwarding.msg_forwarding_service import (
    SignedMessageWithoutReplyForwardingService,
)
from .node_service.node_service import EventualNodeServiceWithoutReply
from .node_service.node_service import ImmediateNodeServiceWithReply
from .node_service.object_action.obj_action_service import (
    EventualObjectActionServiceWithoutReply,
)
from .node_service.object_action.obj_action_service import (
    ImmediateObjectActionServiceWithReply,
)
from .node_service.object_action.obj_action_service import (
    ImmediateObjectActionServiceWithoutReply,
)
from .node_service.object_search.obj_search_service import ImmediateObjectSearchService
from .node_service.object_search_permission_update.obj_search_permission_service import (
    ImmediateObjectSearchPermissionUpdateService,
)
from .node_service.resolve_pointer_type.resolve_pointer_type_service import (
    ResolvePointerTypeService,
)
from .node_service.testing_services.repr_service import ReprService
from .node_service.vpn.vpn_messages import VPNRegisterMessage
from .node_table import Base
from .node_table.node import Node as NodeRow

# this generic type for Client bound by Client
ClientT = TypeVar("ClientT", bound=Client)


# TODO: Move but right now import loop prevents importing from the RequestMessage
class DuplicateRequestException(Exception):
    pass


class Node(AbstractNode):

    """
    Basic class for a syft node behavior, explicit purpose node will
    inherit this class (e.g., Device, Domain, Network, and VirtualMachine).

    Each node is identified by an id of type ID and a name of type string.
    """

    client_type = ClientT
    child_type_client_type = ClientT

    ChildT = TypeVar("ChildT", bound="Node")
    child_type = ChildT

    signing_key: Optional[SigningKey]
    verify_key: Optional[VerifyKey]

    def __init__(
        self,
        name: Optional[str] = None,
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
        TableBase: Any = None,
        db_engine: Any = None,
        store_type: type = RedisStore,
        settings: Optional[BaseSettings] = None,
    ):

        # The node has a name - it exists purely to help the
        # end user have some idea about what this node is in a human
        # readable form. It is not guaranteed to be unique (or to
        # really be anything for that matter).
        super().__init__(
            name=name, network=network, domain=domain, device=device, vm=vm
        )

        self.settings = settings

        # TableBase is the base class from which all ORM classes must inherit
        # If one isn't provided then we can simply make one.
        if TableBase is None:
            TableBase = declarative_base()

        # If not provided a session connecting us to the database, let's just
        # initialize a database in memory
        if db_engine is None:
            db_engine = create_engine("sqlite://", echo=False)
            Base.metadata.create_all(db_engine)  # type: ignore

        # cache these variables on self
        self.TableBase = TableBase
        self.db_engine = db_engine
        # self.db = db
        # self.session = db

        # launch the tables in the database
        # Tudor: experimental
        # self.TableBase.metadata.create_all(engine)

        # Any object that needs to be stored on a node is stored here
        # More specifically, all collections of objects are found here
        # There should be NO COLLECTIONS stored as attributes directly
        # on a Node if there is a chance that the collections could
        # become quite numerous (or otherwise fill up RAM).
        # self.store is the elastic memory.

        self.store = store_type(db=self.db_engine, settings=settings)
        self.setup = SetupManager(database=self.db_engine)

        # We need to register all the services once a node is created
        # On the off chance someone forgot to do this (super unlikely)
        # this flag exists to check it.
        self.services_registered = False

        # In order to be able to write generic services (in .service)
        # which can work for all node types, sometimes we need to have
        # a reference to what node type this node is. This attribute
        # provides that ability.
        self.node_type = type(self).__name__
        # ABOUT SERVICES AND MESSAGES

        # Each service corresponds to one or more message types which
        # the service processes. There are two kinds of messages, those
        # which require a reply and those which do not. Thus, there
        # are two kinds of services, service which generate a reply
        # and services which do not. It's important to distinguish
        # between them because:
        #
        # 1) services which do not return a reply
        # can typically be run on a more flexible time-table, whereas
        # services which require a reply often care about the latency
        # of the reply.
        #
        # 2) Services which do not return a reply aren't likely to leak
        # any information because no information is leaving. Thus, our
        # privacy/security concerns are more concentrated within service
        # which reply with some amount of information.

        # for messages which need a reply, this uses the type
        # of the message to look up the service which
        # addresses that message.
        self.immediate_msg_with_reply_router: Dict[
            Type[ImmediateSyftMessageWithReply], ImmediateNodeServiceWithReply
        ] = {}

        # for messages which don't lead to a reply, this uses
        # the type of the message to look up the service
        # which addresses that message.
        self.immediate_msg_without_reply_router: Dict[
            Type[ImmediateSyftMessageWithoutReply], Any
        ] = {}

        # for messages which don't need to be run right now
        # and will not generate a reply.
        self.eventual_msg_without_reply_router: Dict[
            Type[EventualSyftMessageWithoutReply], EventualNodeServiceWithoutReply
        ] = {}

        # This is the list of services which all node support.
        # You can read more about them by reading their respective
        # class documentation.

        # TODO: Support ImmediateNodeServiceWithoutReply Parent Class
        # for services which run immediately and do not return a reply
        self.immediate_services_without_reply: List[Any] = []
        self.immediate_services_without_reply.append(ReprService)
        self.immediate_services_without_reply.append(HeritageUpdateService)
        self.immediate_services_without_reply.append(ChildNodeLifecycleService)
        self.immediate_services_without_reply.append(
            ImmediateObjectActionServiceWithoutReply
        )
        self.immediate_services_without_reply.append(
            ImmediateObjectSearchPermissionUpdateService
        )

        # TODO: Support ImmediateNodeServiceWithReply Parent Class
        # for services which run immediately and return a reply
        self.immediate_services_with_reply: List[Any] = []
        self.immediate_services_with_reply.append(ImmediateObjectActionServiceWithReply)
        self.immediate_services_with_reply.append(ImmediateObjectSearchService)
        self.immediate_services_with_reply.append(GetReprService)
        self.immediate_services_with_reply.append(ResolvePointerTypeService)

        # for services which can run at a later time and do not return a reply
        self.eventual_services_without_reply = list()
        self.eventual_services_without_reply.append(
            EventualObjectActionServiceWithoutReply
        )

        # This is a special service which cannot be listed in any
        # of the other services because it handles messages of all types.
        # Thus, it does not live in a message router since
        # routers only exist to decide which messages go to which
        # services, and they require that every message only correspond
        # to only one service type. If we have more messages like
        # these we'll make a special category for "services that
        # all messages are applied to" but for now this will do.

        self.signed_message_with_reply_forwarding_service = (
            SignedMessageWithReplyForwardingService()
        )

        self.signed_message_without_reply_forwarding_service = (
            SignedMessageWithoutReplyForwardingService()
        )

        self.allowed_unsigned_messages = []
        self.allowed_unsigned_messages.append(VPNRegisterMessage)

        # now we need to load the relevant frameworks onto the node
        self.lib_ast = lib_ast
        # The node needs to sign messages that it sends so that recipients know that it
        # comes from the node. In order to do that, the node needs to generate keys
        # for itself to sign and verify with.

        # create a signing key if one isn't provided
        if signing_key is None:
            self.signing_key = SigningKey.generate()
        else:
            self.signing_key = signing_key

        # if verify key isn't provided, get verify key from signing key
        if verify_key is None:
            self.verify_key = self.signing_key.verify_key
        else:
            self.verify_key = verify_key

        # PERMISSION REGISTRY:
        self.root_verify_key = self.verify_key  # TODO: CHANGE
        self.guest_verify_key_registry = set()
        self.admin_verify_key_registry = set()
        self.cpl_ofcr_verify_key_registry = set()
        self.peer_route_clients: Dict[UID, Dict[str, Dict[str, Client]]] = {}
        # TODO: remove hacky signaling_msgs when SyftMessages become Storable.
        self.signaling_msgs = {}

        # For logging the number of messages received
        self.message_counter = 0

    def post_init(self) -> None:
        debug(f"> Creating {self.pprint}")

    def set_node_uid(self) -> None:
        try:
            setup = self.setup.first()
            # if its empty it will be set during CreateInitialSetUpMessage
            if setup.node_id != "":
                try:
                    node_id = UID.from_string(setup.node_id)
                except Exception as e:
                    error(f"Invalid Node UID in Setup Table. {setup.node_id}")
                    raise e

                location = SpecificLocation(name=setup.domain_name, id=node_id)
                # TODO: Fix with proper isinstance when the class will import
                if type(self).__name__ == "Domain":
                    self.domain = location
                elif type(self).__name__ == "Network":
                    self.network = location
                info(f"Finished setting Node UID. {location}")
        except Exception:
            info("Setup hasnt run yet so ignoring set_node_uid")
            pass

    @property
    def icon(self) -> str:
        return "📍"

    def get_client(
        self,
        routes: Optional[List[Route]] = None,
        signing_key: Optional[SigningKey] = None,
    ) -> ClientT:
        if not routes:
            conn_client = create_virtual_connection(node=self)
            solo = SoloRoute(destination=self.target_id, connection=conn_client)
            # inject name
            setattr(
                solo,
                "name",
                f"Route ({self.name} <-> {self.name} Client)",
            )
            routes = [solo]

        return self.client_type(  # type: ignore
            name=self.name,
            routes=routes,
            network=self.network,
            domain=self.domain,
            device=self.device,
            vm=self.vm,
            signing_key=signing_key,  # If no signing_key is passed, the client generates one.
            verify_key=None,  # DO NOT PASS IN A VERIFY KEY!!! The client generates one.
        )

    def get_root_client(self, routes: Optional[List[Route]] = None) -> ClientT:
        client: ClientT = self.get_client(routes=routes)
        self.root_verify_key = client.verify_key
        return client

    def get_metadata_for_client(self) -> Metadata:
        return Metadata(
            name=self.name if self.name else "",
            id=self.id,
            node=self.target_id,
            node_type=str(type(self).__name__),
            version=str(sy.__version__),
        )

    def add_peer_routes(self, peer: NodeRow) -> None:
        try:
            routes = self.node_route.query(node_id=peer.id)  # type: ignore
            for route in routes:
                self.add_route(
                    node_id=UID.from_string(value=peer.node_uid),
                    node_name=peer.node_name,
                    host_or_ip=route.host_or_ip,
                    is_vpn=route.is_vpn,
                )
        except Exception as e:
            error(f"Failed to add route to peer {peer}. {e}")

    def reload_peer_clients(self) -> None:
        peers = self.node.all()  # type: ignore
        for peer in peers:
            self.add_peer_routes(peer=peer)
        debug("Finished loading all the peer clients", self.peer_route_clients)

    def all_peer_clients(self) -> Dict[UID, List[Client]]:
        # get all the routes for each client and sort by VPN first
        all_clients = {}
        for node_id in self.peer_route_clients.keys():
            all_clients[node_id] = list(
                self.peer_route_clients[node_id]["vpn"].values()
            ) + list(self.peer_route_clients[node_id]["public"].values())

        return all_clients

    def add_route(
        self, node_id: UID, node_name: str, host_or_ip: str, is_vpn: bool
    ) -> None:
        debug(f"Adding route {node_id}, {node_name}, {host_or_ip}, {is_vpn}")
        try:
            vpn_key = "vpn" if is_vpn else "public"
            # make sure the node_id is in the Dict
            node_id_dict: Dict[str, Dict[str, Client]] = {"vpn": {}, "public": {}}
            if node_id in self.peer_route_clients:
                node_id_dict = self.peer_route_clients[node_id]

            if host_or_ip not in node_id_dict[vpn_key]:
                # connect and save the client
                grid_url = GridURL.from_url(host_or_ip)
                client = sy.connect(url=grid_url.with_path("/api/v1"), timeout=0.3)
                node_id_dict[vpn_key][host_or_ip] = client

            self.peer_route_clients[node_id] = node_id_dict
        except Exception as e:
            debug(
                f"Failed to add_route {node_id} {node_name} {host_or_ip} {is_vpn}. {e}"
            )

    def get_peer_client(self, node_id: UID, only_vpn: bool = True) -> Optional[Client]:
        # if we don't have it see if we can get it from the db first
        if node_id not in self.peer_route_clients:
            peer = self.node.first(node_uid=node_id.no_dash)  # type: ignore
            self.add_peer_routes(peer=peer)

        try:
            if node_id in self.peer_route_clients.keys():
                routes = self.peer_route_clients[node_id]
                # if we want VPN only then check there are some
                if only_vpn and "vpn" in routes and len(routes["vpn"]) == 0:
                    # we want VPN only but there are none
                    return None
                elif "vpn" in routes and len(routes["vpn"]) > 0:
                    # if we have VPN lets use it
                    return list(routes["vpn"].values())[0]
                elif "public" in routes and len(routes["public"]) > 0:
                    # we only have public and don't care
                    return list(routes["public"].values())[0]
        except Exception as e:
            error(
                f"Exception while selecting node_id {node_id} from peer_route_clients. "
                f"{self.peer_route_clients}. {e}"
            )

        # there are no routes for this ID
        return None

    @property
    def id(self) -> UID:
        traceback_and_raise(NotImplementedError)

    def message_is_for_me(self, msg: Union[SyftMessage, SignedMessage]) -> bool:
        traceback_and_raise(NotImplementedError)

    def recv_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> SignedImmediateSyftMessageWithoutReply:
        contents = getattr(msg, "message", msg)
        # exceptions can be easily triggered which break any loops
        # so we need to catch them here and respond with a special exception
        # message reply
        try:
            debug(
                f"> Received with Reply {contents.pprint} {contents.id} @ {self.pprint}"
            )
            # try to process message
            response = self.process_message(
                msg=msg, router=self.immediate_msg_with_reply_router
            )

        except Exception as e:
            print(type(e), e)
            error(e)
            public_exception: Exception
            if isinstance(e, AuthorizationException):
                private_log_msg = "An AuthorizationException has been triggered"
                public_exception = e
            else:
                private_log_msg = f"An {type(e)} has been triggered"  # dont send
                public_exception = UnknownPrivateException(
                    "UnknownPrivateException has been triggered."
                )
            try:
                # try printing a useful message
                private_log_msg += f" by {type(contents)} "
                private_log_msg += f"from {contents.reply_to}"  # type: ignore
            except Exception:
                error("Unable to format the private log message")
                pass
            # show the host what the real exception is
            error(private_log_msg)

            # send the public exception back
            response = ExceptionMessage(
                address=contents.reply_to,  # type: ignore
                msg_id_causing_exception=contents.id,
                exception_type=type(public_exception),
                exception_msg=str(public_exception),
            )

        # maybe I shouldn't have created process_message because it screws up
        # all the type inference.
        res_msg = response.sign(signing_key=self.signing_key)  # type: ignore
        output = (
            f"> {self.pprint} Signing {res_msg.pprint} with "
            + f"{self.key_emoji(key=self.signing_key.verify_key)}"  # type: ignore
        )
        debug(output)
        return res_msg

    def recv_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        contents = getattr(msg, "message", msg)
        if contents:
            debug(
                f"> Received without Reply {contents.pprint} {contents.id} @ {self.pprint}"
            )

        self.process_message(msg=msg, router=self.immediate_msg_without_reply_router)
        try:
            pass
        except Exception as e:
            error(f"Exception processing {contents}. {e}")
            # public_exception: Exception
            if isinstance(e, DuplicateRequestException):
                private_log_msg = "An DuplicateRequestException has been triggered"
                # public_exception = e
            else:
                private_log_msg = f"An {type(e)} has been triggered"  # dont send
                # public_exception = UnknownPrivateException(
                #     "UnknownPrivateException has been triggered."
                # )
            try:
                # try printing a useful message
                private_log_msg += f" by {type(contents)} "
                private_log_msg += f"from {contents.reply_to}"  # type: ignore
            except Exception:
                error("Unable to format the private log message")
                pass
            # show the host what the real exception is
            error(private_log_msg)

            # we still want to raise for now due to certain exceptions we expect
            # in tests
            if not isinstance(e, DuplicateRequestException):
                error(e)
                # TODO: A lot of tests are depending on this raise which seems bad
                traceback_and_raise(e)

            # TODO: finish code to send ExceptionMessage back
            # if isinstance(e, DuplicateRequestException):
            #     # we have a reply_to
            #     # send the public exception back
            #     response = ExceptionMessage(
            #         address=msg.message.reply_to,  # type: ignore
            #         msg_id_causing_exception=msg.message.id,
            #         exception_type=type(public_exception),
            #         exception_msg=str(public_exception),
            #     )
            #     res_msg = response.sign(signing_key=self.signing_key)  # type: ignore
            #     self.client.send_immediate_msg_with_reply.process(
            #         node=self,
            #         msg=res_msg,
            #     )
        return None

    def recv_eventual_msg_without_reply(
        self, msg: SignedEventualSyftMessageWithoutReply
    ) -> None:
        self.process_message(msg=msg, router=self.eventual_msg_without_reply_router)

    # TODO: Add SignedEventualSyftMessageWithoutReply and others
    def process_message(
        self, msg: SignedMessage, router: dict
    ) -> Union[SyftMessage, None]:
        self.message_counter += 1
        try:
            contents = getattr(
                msg, "message", msg
            )  # in the event the message is unsigned
            debug(f"> Processing 📨 {msg.pprint} @ {self.pprint} {contents}")
            if self.message_is_for_me(msg=msg):
                debug(
                    f"> Recipient Found {msg.pprint}{msg.address.target_emoji()} == {self.pprint}"
                )

                # only a small number of messages are allowed to be unsigned otherwise
                # they need to be valid
                if type(msg) not in self.allowed_unsigned_messages and not msg.is_valid:  # type: ignore
                    error(f"Message is not valid. {msg}")
                    traceback_and_raise(Exception("Message is not valid."))

                # Process Message here
                try:  # we use try/except here because it's marginally faster in Python
                    service = router[type(contents)]
                except KeyError as e:
                    log = (
                        f"The node {self.id} of type {type(self)} cannot process messages of type "
                        + f"{type(contents)} because there is no service running to process it."
                        + f"{e}"
                    )
                    error(log)
                    self.ensure_services_have_been_registered_error_if_not()
                    traceback_and_raise(KeyError(log))

                if type(msg) in self.allowed_unsigned_messages:  # type: ignore
                    result = service.process(node=self, msg=contents, verify_key=None)
                else:
                    result = service.process(
                        node=self,
                        msg=contents,
                        verify_key=msg.verify_key,
                    )
                return result

            else:
                debug(
                    f"> Recipient Not Found ↪️ {msg.pprint}{msg.address.target_emoji()} != {self.pprint}"
                )
                # Forward message onwards
                if issubclass(type(msg), SignedImmediateSyftMessageWithReply):
                    return self.signed_message_with_reply_forwarding_service.process(
                        node=self,
                        msg=msg,  # type: ignore
                    )
                if issubclass(type(msg), SignedImmediateSyftMessageWithoutReply):
                    return self.signed_message_without_reply_forwarding_service.process(
                        node=self,
                        msg=msg,  # type: ignore
                    )
        except Exception as e:
            error(e)
            raise e
        return None

    def ensure_services_have_been_registered_error_if_not(self) -> None:
        if not self.services_registered:
            traceback_and_raise(
                Exception(
                    "Please call _register_services on node. This seems to have"
                    "been skipped for some reason."
                )
            )

    def _register_services(self) -> None:
        """In this method, we set each message type to the appropriate
        service for this node. It's important to note that one message type
        cannot map to multiple services on any given node type. If you want to
        send information to a different service, create a new message type for that
        service. Put another way, a service can have multiple message types which
        correspond to it, but each message type can only have one service (per node
        subclass) which corresponds to it."""

        for isr in self.immediate_services_with_reply:
            # Create a single instance of the service to cache in the router corresponding
            # to one or more message types.
            isr_instance = isr()
            for handler_type in isr.message_handler_types():
                # for each explicitly supported type, add it to the router
                self.immediate_msg_with_reply_router[handler_type] = isr_instance

                # for all sub-classes of the explicitly supported type, add them
                # to the router as well.
                for handler_type_subclass in get_subclasses(obj_type=handler_type):
                    self.immediate_msg_with_reply_router[
                        handler_type_subclass
                    ] = isr_instance

        for iswr in self.immediate_services_without_reply:
            # Create a single instance of the service to cache in the router corresponding
            # to one or more message types.
            iswr_instance = iswr()
            for handler_type in iswr.message_handler_types():

                # for each explicitly supported type, add it to the router
                self.immediate_msg_without_reply_router[handler_type] = iswr_instance

                # for all sub-classes of the explicitly supported type, add them
                # to the router as well.
                for handler_type_subclass in get_subclasses(obj_type=handler_type):
                    self.immediate_msg_without_reply_router[
                        handler_type_subclass
                    ] = iswr_instance

        for eswr in self.eventual_services_without_reply:
            # Create a single instance of the service to cache in the router corresponding
            # to one or more message types.
            eswr_instance = eswr()
            for handler_type in eswr.message_handler_types():

                # for each explicitly supported type, add it to the router
                self.eventual_msg_without_reply_router[handler_type] = eswr_instance

                # for all sub-classes of the explicitly supported type, add them
                # to the router as well.
                for handler_type_subclass in get_subclasses(obj_type=handler_type):
                    self.eventual_msg_without_reply_router[
                        handler_type_subclass
                    ] = eswr_instance

        # Set the services_registered flag to true so that we know that all services
        # have been properly registered. This mostly exists because someone might
        # accidentally delete (forget to call) this method inside the __init__ function
        # of a sub-class of Node.
        self.services_registered = True

    def __repr__(self) -> str:
        no_dash = str(self.id).replace("-", "")
        return f"{self.node_type}: {self.name}: {no_dash}"



"""This file defines classes and methods which are used to manage database queries on the SyftUser table."""

# stdlib
from datetime import datetime
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from bcrypt import checkpw
from bcrypt import gensalt
from bcrypt import hashpw
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Query
from sqlalchemy.orm import sessionmaker

# relative
from ..exceptions import InvalidCredentialsError
from ..exceptions import UserNotFoundError
from ..node_table.pdf import PDFObject
from ..node_table.roles import Role
from ..node_table.user import SyftUser
from ..node_table.user import UserApplication
from .constants import UserApplicationStatus
from .database_manager import DatabaseManager
from .role_manager import RoleManager


class UserManager(DatabaseManager):
    """Class to manage user database actions."""

    schema = SyftUser

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=UserManager.schema, db=database)
        self.roles = RoleManager(database)

    @property
    def common_users(self) -> list:
        """Return users having the common role access."""
        common_users: List[SyftUser] = []
        for role in self.roles.common_roles:
            common_users = common_users + list(super().query(role=role.id))

        return common_users

    @property
    def org_users(self) -> list:
        """Return all the users in the organization."""
        org_users: List[SyftUser] = []
        for role in self.roles.org_roles:
            org_users = org_users + list(super().query(role=role.id))
        return org_users

    def create_user_application(
        self,
        name: str,
        email: str,
        password: str,
        daa_pdf: Optional[bytes],
        institution: Optional[str] = "",
        website: Optional[str] = "",
        budget: Optional[float] = 0.0,
    ) -> int:
        """Stores the information of the application submitted by the user.

        Args:
            name (str): name of the user.
            email (str): email of the user.
            password (str): password of the user.
            daa_pdf (Optional[bytes]): data access agreement.
            institution (Optional[str], optional): name of the institution to which the user belongs. Defaults to "".
            website (Optional[str], optional): website link of the institution. Defaults to "".
            budget (Optional[float], optional): privacy budget allocated to the user. Defaults to 0.0.

        Returns:
            int: Id of the application of the user.
        """

        salt, hashed = self.__salt_and_hash_password(password, 12)
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        _pdf_obj = PDFObject(binary=daa_pdf)
        session_local.add(_pdf_obj)
        session_local.commit()
        session_local.flush()
        session_local.refresh(_pdf_obj)

        _obj = UserApplication(
            name=name,
            email=email,
            salt=salt,
            hashed_password=hashed,
            daa_pdf=_pdf_obj.id,
            institution=institution,
            website=website,
            budget=budget,
        )
        session_local.add(_pdf_obj)
        session_local.add(_obj)
        session_local.commit()
        _obj_id = _obj.id
        session_local.close()
        return _obj_id

    def get_all_applicant(self) -> List[UserApplication]:
        """Returns the application data of all the applicants in the database.

        Returns:
            List[UserApplication]: All user applications.
        """
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        result = list(session_local.query(UserApplication).all())
        session_local.close()
        return result

    def process_user_application(
        self, candidate_id: int, status: str, verify_key: VerifyKey
    ) -> None:
        """Process the application for the given candidate.

        If the application of the user was accepted, then register the user
        and its details in the database. Finally update the application status
        for the given user/candidate in the database.

        Args:
            candidate_id (int): user id of the candidate.
            status (str): application status.
            verify_key (VerifyKey): public digital signature of the user.
        """
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        candidate = (
            session_local.query(UserApplication).filter_by(id=candidate_id).first()
        )
        session_local.close()

        if (
            status == UserApplicationStatus.ACCEPTED.value
        ):  # If application was accepted
            # Generate a new signing key
            _private_key = SigningKey.generate()

            encoded_pk = _private_key.encode(encoder=HexEncoder).decode("utf-8")
            encoded_vk = _private_key.verify_key.encode(encoder=HexEncoder).decode(
                "utf-8"
            )
            added_by = self.get_user(verify_key).name  # type: ignore

            # Register the user in the database
            self.register(
                name=candidate.name,
                email=candidate.email,
                role=self.roles.ds_role.id,
                budget=candidate.budget,
                private_key=encoded_pk,
                verify_key=encoded_vk,
                hashed_password=candidate.hashed_password,
                salt=candidate.salt,
                daa_pdf=candidate.daa_pdf,
                added_by=added_by,
                institution=candidate.institution,
                website=candidate.website,
                created_at=datetime.now(),
            )
        else:
            status = UserApplicationStatus.REJECTED.value

        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        candidate = (
            session_local.query(UserApplication).filter_by(id=candidate_id).first()
        )
        candidate.status = status
        session_local.flush()
        session_local.commit()
        session_local.close()

    def signup(
        self,
        name: str,
        email: str,
        password: str,
        budget: float,
        role: int,
        private_key: str,
        verify_key: str,
    ) -> SyftUser:
        """Registers a user in the database, when they signup on a domain.

        Args:
            name (str): name of the user.
            email (str): email of the user.
            password (str): password set by the user.
            budget (float): privacy budget alloted to the user.
            role (int): role of the user when they signup on the domain.
            private_key (str): private digital signature of the user.
            verify_key (str): public digital signature of the user.

        Returns:
            SyftUser: the registered user object.
        """
        salt, hashed = self.__salt_and_hash_password(password, 12)
        return self.register(
            name=name,
            email=email,
            role=role,
            budget=budget,
            private_key=private_key,
            verify_key=verify_key,
            hashed_password=hashed,
            salt=salt,
            created_at=datetime.now(),
        )

    def query(self, **kwargs: Any) -> Query:
        results = super().query(**kwargs)
        return results

    def first(self, **kwargs: Any) -> SyftUser:
        result = super().first(**kwargs)
        if not result:
            raise UserNotFoundError
        return result

    def login(self, email: str, password: str) -> SyftUser:
        """Returns the user object for the given the email and password.

        Args:
            email (str): email of the user.
            password (str): password of the user.

        Returns:
            SyftUser: user object for the given email and password.
        """
        return self.__login_validation(email, password)

    def set(  # nosec
        self,
        user_id: str,
        email: str = "",
        password: str = "",
        role: int = 0,
        name: str = "",
        website: str = "",
        institution: str = "",
        budget: float = 0.0,
    ) -> None:
        """Updates the information for the given user id.

        Args:
            user_id (str): unique id of the user in the database.
            email (str, optional): email of the user. Defaults to "".
            password (str, optional): password of the user. Defaults to "".
            role (int, optional): role of the user. Defaults to 0.
            name (str, optional): name of the user. Defaults to "".
            website (str, optional): website of the institution of the user. Defaults to "".
            institution (str, optional): name of the institution of the user. Defaults to "".
            budget (float, optional): privacy budget allocated to the user. Defaults to 0.0.

        Raises:
            UserNotFoundError: Raised when a user does not exits for the given user id.
            Exception: Raised when an invalid argument/property is passed.
        """
        if not self.contain(id=user_id):
            raise UserNotFoundError

        key: str
        value: Union[str, int, float]

        if email:
            key = "email"
            value = email
        elif password:
            salt, hashed = self.__salt_and_hash_password(password, 12)
            self.modify({"id": user_id}, {"salt": salt, "hashed_password": hashed})
            return
        elif role != 0:
            key = "role"
            value = role
        elif name:
            key = "name"
            value = name
        elif budget:
            key = "budget"
            value = budget
        elif website:
            key = "website"
            value = website
        elif budget:
            key = "budget"
            value = budget
        elif institution:
            key = "institution"
            value = institution
        else:
            raise Exception

        self.modify({"id": user_id}, {key: value})

    def can_create_users(self, verify_key: VerifyKey) -> bool:
        """Checks if a user has permissions to create new users."""
        try:
            return self.role(verify_key=verify_key).can_create_users
        except UserNotFoundError:
            return False

    def can_upload_data(self, verify_key: VerifyKey) -> bool:
        """Checks if a user has permissions to upload data to the node."""
        try:
            return self.role(verify_key=verify_key).can_upload_data
        except UserNotFoundError:
            return False

    def can_triage_requests(self, verify_key: VerifyKey) -> bool:
        """Checks if a user has permissions to triage requests."""
        try:
            return self.role(verify_key=verify_key).can_triage_data_requests
        except UserNotFoundError:
            return False

    def can_manage_infrastructure(self, verify_key: VerifyKey) -> bool:
        """Checks if a user has permissions to manage the deployed infrastructure."""
        try:
            return self.role(verify_key=verify_key).can_manage_infrastructure
        except UserNotFoundError:
            return False

    def can_edit_roles(self, verify_key: VerifyKey) -> bool:
        """Checks if a user has permission to edit roles of other users."""
        try:
            return self.role(verify_key=verify_key).can_edit_roles
        except UserNotFoundError:
            return False

    def role(self, verify_key: VerifyKey) -> Role:
        """Returns the role of the given user."""
        user = self.get_user(verify_key)
        if not user:
            raise UserNotFoundError
        return self.roles.first(id=user.role)

    def get_user(self, verify_key: VerifyKey) -> Optional[SyftUser]:
        """Returns the user for the given public digital signature."""
        return self.first(
            verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
        )

    def __login_validation(self, email: str, password: str) -> SyftUser:
        """Validates and returns the user object for the given credentials.

        Args:
            email (str): email of the user.
            password (str): password of the user.

        Raises:
            UserNotFoundError: Raised if the user does not exist for the email.
            InvalidCredentialsError: Raised if either the password or email is incorrect.

        Returns:
            SyftUser: Returns the user for the given credentials.
        """
        try:
            user = self.first(email=email)
            if not user:
                raise UserNotFoundError

            hashed = user.hashed_password.encode("UTF-8")
            salt = user.salt.encode("UTF-8")
            bytes_pass = password.encode("UTF-8")

            if checkpw(bytes_pass, salt + hashed):
                return user
            else:
                raise InvalidCredentialsError
        except UserNotFoundError:
            raise InvalidCredentialsError

    def __salt_and_hash_password(self, password: str, rounds: int) -> Tuple[str, str]:
        bytes_pass = password.encode("UTF-8")
        salt = gensalt(rounds=rounds)
        salt_len = len(salt)
        hashed = hashpw(bytes_pass, salt)
        hashed = hashed[salt_len:]
        hashed = hashed.decode("UTF-8")
        salt = salt.decode("UTF-8")
        return salt, hashed



""" A set of util methods used by the syft.lib submodule. """
# stdlib
from typing import Callable
from typing import Union as TypeUnion

# relative
from ..ast.globals import Globals
from ..core.node.abstract.node import AbstractNodeClient

# this gets called on global ast as well as clients
# anything which wants to have its ast updated and has an add_attr method


def generic_update_ast(
    lib_name: str,
    create_ast: Callable,
    ast_or_client: TypeUnion[Globals, AbstractNodeClient],
) -> None:
    """
    Update the abstract syntax tree data structure used by the Globals or AbstractNodeClient entities.

    Args:
        lib_name (str): Library name to update ast.
        create_ast (Callable): AST generation function for the given library.
        ast_or_client (Union[Globals, AbstractNodeClient]): Entity to update ast.

    Raises:
        ValueError: raises a ValueError exception if ast_or_client isn't a Globals or AbstractNodeClient instance.
    """
    if isinstance(ast_or_client, Globals):
        ast = ast_or_client
        new_lib_ast = create_ast(None)
        ast.add_attr(attr_name=lib_name, attr=new_lib_ast.attrs[lib_name])
    elif isinstance(ast_or_client, AbstractNodeClient):
        client = ast_or_client
        new_lib_ast = create_ast(client)
        client.lib_ast.attrs[lib_name] = new_lib_ast.attrs[lib_name]
        setattr(client, lib_name, new_lib_ast.attrs[lib_name])
    else:
        raise ValueError(
            f"Expected param of type (Globals, AbstractNodeClient), but got {type(ast_or_client)}"
        )


def full_name_with_qualname(klass: type) -> str:
    """Returns the klass module name + klass qualname."""
    return f"{klass.__module__}.{klass.__qualname__}"


def full_name_with_name(klass: type) -> str:
    """Returns the klass module name + klass name."""
    return f"{klass.__module__}.{klass.__name__}"



# third party
import torch

# relative
from ...core.common.serde.serializable import serializable
from ...proto.lib.torch.size_pb2 import Size as TorchSize_PB


def protobuf_torch_size_serializer(torch_size: torch.Size) -> TorchSize_PB:
    serialized_size = TorchSize_PB(data=torch_size)
    return serialized_size


def protobuf_torch_size_deserializer(proto_size: TorchSize_PB) -> torch.Size:
    torch_size = torch.Size(proto_size.data)
    return torch_size


serializable(generate_wrapper=True)(
    wrapped_type=torch.Size,
    import_path="torch.Size",
    protobuf_scheme=TorchSize_PB,
    type_object2proto=protobuf_torch_size_serializer,
    type_proto2object=protobuf_torch_size_deserializer,
)



# stdlib
from typing import Any
from typing import Optional
from typing import Union

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from pydantic import BaseSettings
from typing_extensions import final

# relative
from ....logger import critical
from ....logger import traceback_and_raise
from ...common.message import SignedMessage
from ...common.message import SyftMessage
from ...common.uid import UID
from ...io.location import Location
from ...io.location import SpecificLocation
from ..common.node import Node
from ..common.node_manager.dict_store import DictStore
from .client import VirtualMachineClient


@final
class VirtualMachine(Node):
    client_type = VirtualMachineClient
    vm: SpecificLocation  # redefine the type of self.vm to not be optional
    signing_key: Optional[SigningKey]
    verify_key: Optional[VerifyKey]
    child_type_client_type = None

    def __init__(
        self,
        *,  # Trasterisk
        name: Optional[str] = None,
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: SpecificLocation = SpecificLocation(),
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
        store_type: type = DictStore,
        settings: Optional[BaseSettings] = None,
    ):
        super().__init__(
            name=name,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
            store_type=store_type,
        )

        # specific location with name
        self.vm = SpecificLocation(name=self.name)

        # relative
        from ..common.node_service.vm_request_service.vm_service import (
            VMRequestAnswerService,
        )

        self.immediate_services_with_reply.append(VMRequestAnswerService)
        # All node subclasses have to call this at the end of their __init__
        self._register_services()
        self.post_init()

    @property
    def icon(self) -> str:
        return "🍰"

    @property
    def id(self) -> UID:
        return self.vm.id

    def message_is_for_me(self, msg: Union[SyftMessage, SignedMessage]) -> bool:
        # this needs to be defensive by checking vm_id NOT vm.id or it breaks
        try:
            return msg.address.vm_id == self.id
        except Exception as e:
            critical(f"Error checking if {msg.pprint} is for me on {self.pprint}. {e}")
            return False

    def _register_frameworks(self) -> None:
        traceback_and_raise(NotImplementedError)
        # TODO: it doesn't at the moment but it needs to in the future,
        #  mostly because nodes should be able to choose waht framweorks they
        #  want to support (and more importantly what versions of those frameworks
        #  they want to support).
        # QUESTION: Does this exist?
        # from ....lib import supported_frameworks
        # for fw in supported_frameworks:
        #     for name, ast in fw.ast.attrs.items():
        #         if name in self.frameworks.attrs:
        #             traceback_and_raise(KeyError(
        #                 "Framework already imported. Why are you importing it twice?"
        #             ))
        #         self.frameworks.attrs[name] = ast

    def __hash__(self) -> int:
        return hash(self.vm.id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, VirtualMachine):
            return False

        if self.vm.id != other.vm.id:
            return False

        return True



# stdlib
from typing import List
from typing import Optional

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from typing_extensions import final

# relative
from ...common.uid import UID
from ...io.location import Location
from ...io.location.specific import SpecificLocation
from ...io.route import Route
from ..common.client import Client


@final
class VirtualMachineClient(Client):

    vm: SpecificLocation  # redefine the type of self.vm to not be optional

    def __init__(
        self,
        name: Optional[str],
        routes: List[Route],
        vm: SpecificLocation,
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
    ):
        super().__init__(
            name=name,
            routes=routes,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
        )

        self.post_init()

    @property
    def id(self) -> UID:
        return self.vm.id

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.name}>"



# stdlib
from uuid import UUID

# third party
from numpy import ndarray

# relative
from .....core.common.serde.serializable import serializable
from ..tensor import AutogradTensor
from .op import Op


@serializable(recursive_serde=True)
class ReshapeOp(Op):
    """Multiplication operation with 2 tensors"""

    __attr_allowlist__ = ["x", "shape", "backward_shape"]

    def forward(self, x: AutogradTensor, *shape: tuple) -> AutogradTensor:  # type: ignore
        self.x = x
        self.shape = shape
        self.backward_shape = self.x.shape

        return AutogradTensor(x.child.reshape(*shape), requires_grad=x.requires_grad)

    def _backward(self, grad: ndarray, backprop_id: UUID) -> None:

        if self.x.requires_grad:
            self.x.add_grad(grad.reshape(*self.backward_shape))

            if self.x.grad_fn:
                self.x.backward(backprop_id=backprop_id)



# stdlib
from typing import Any
from typing import List
from typing import Optional

# relative
from .uid import UID


class AbstractPointer:
    def __init__(
        self,
        client: Any,
        id_at_location: Optional[UID] = None,
        tags: Optional[List[str]] = None,
        description: str = "",
    ):
        if id_at_location is None:
            id_at_location = UID()

        if tags is None:
            tags = []

        self.client = client
        self.id_at_location = id_at_location
        self.tags = tags
        self.description = description
        self.gc_enabled = True

        self.is_enum = False



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

# third party
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

# relative
from ..node_table import Base


class DatabaseManager:
    def __init__(self, schema: Type[Base], db: Engine) -> None:
        self._schema = schema
        self.db = db

    def register(self, **kwargs: Any) -> Any:
        """Register a new object into the database.

        Args:
            parameters : List of object parameters.
        Returns:
            object: Database Object
        """
        _obj = self._schema(**kwargs)
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        session_local.add(_obj)
        session_local.commit()
        return _obj

    def query(self, **kwargs: Any) -> List[Any]:
        """Query db objects filtering by parameters
        Args:
            parameters : List of parameters used to filter.
        """
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        objects = session_local.query(self._schema).filter_by(**kwargs).all()
        session_local.close()
        return objects

    def first(self, **kwargs: Any) -> Optional[Any]:
        """Query db objects filtering by parameters
        Args:
            parameters : List of parameters used to filter.
        """
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        objects = session_local.query(self._schema).filter_by(**kwargs).first()
        session_local.close()
        return objects

    def last(self, **kwargs: Any) -> Optional[Any]:
        """Query and return the last occurrence.

        Args:
            parameters: List of parameters used to filter.
        Return:
            obj: Last object instance.
        """
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        obj = session_local.query(self._schema).filter_by(**kwargs).all()[-1]
        session_local.close()
        return obj

    def all(self) -> List[Any]:
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        result = list(session_local.query(self._schema).all())
        session_local.close()
        return result

    def delete(self, **kwargs: Any) -> None:
        """Delete an object from the database.

        Args:
            parameters: Parameters used to filter the object.
        """
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        session_local.query(self._schema).filter_by(**kwargs).delete()
        session_local.commit()
        session_local.close()

    def modify(self, query: Dict[Any, Any], values: Dict[Any, Any]) -> None:
        """Modifies one or many records."""
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        session_local.query(self._schema).filter_by(**query).update(values)
        session_local.commit()
        session_local.close()

    def contain(self, **kwargs: Any) -> bool:
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        objects = session_local.query(self._schema).filter_by(**kwargs).all()
        session_local.close()
        return len(objects) != 0

    def __len__(self) -> int:
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        result = session_local.query(self._schema).count()
        session_local.close()
        return result

    def clear(self) -> None:
        local_session = sessionmaker(bind=self.db)()
        local_session.query(self._schema).delete()
        local_session.commit()
        local_session.close()



# future
from __future__ import annotations

# stdlib
import operator
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np
import torch as th

# syft absolute
import syft as sy

# relative
from ... import lib
from ...ast.klass import pointerize_args_and_kwargs
from ...util import inherit_tags
from ..common.serde.capnp import CapnpModule
from ..common.serde.capnp import chunk_bytes
from ..common.serde.capnp import combine_bytes
from ..common.serde.capnp import get_capnp_schema
from ..common.serde.capnp import serde_magic_header
from ..common.serde.serializable import serializable
from ..common.uid import UID
from ..node.abstract.node import AbstractNodeClient
from ..node.common.action.run_class_method_action import RunClassMethodAction
from ..pointer.pointer import Pointer
from .ancestors import AutogradTensorAncestor
from .ancestors import PhiTensorAncestor
from .fixed_precision_tensor_ancestor import FixedPrecisionTensorAncestor
from .passthrough import PassthroughTensor  # type: ignore
from .smpc import utils
from .smpc.mpc_tensor import MPCTensor


class TensorPointer(Pointer):

    # Must set these at class init time despite
    # the fact that klass.Class tries to override them (unsuccessfully)
    __name__ = "TensorPointer"
    __module__ = "syft.core.tensor.tensor"

    def __init__(
        self,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
        public_shape: Optional[Tuple[int, ...]] = None,
        public_dtype: Optional[Union[str, np.dtype]] = None,
    ):

        super().__init__(
            client=client,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
        )

        self.public_shape = public_shape

        if isinstance(public_dtype, str):
            self.public_dtype = np.dtype(public_dtype)
        else:
            self.public_dtype = public_dtype

    def share(self, *parties: Tuple[AbstractNodeClient, ...]) -> MPCTensor:
        all_parties = list(parties) + [self.client]
        ring_size = utils.TYPE_TO_RING_SIZE.get(self.public_dtype, None)
        self_mpc = MPCTensor(
            secret=self,
            shape=self.public_shape,
            parties=all_parties,
            ring_size=ring_size,
        )
        return self_mpc

    def _apply_tensor_op(self, other: Any, op_str: str) -> Any:
        # we want to get the return type which matches the attr_path_and_name
        # so we ask lib_ast for the return type name that matches out
        # attr_path_and_name and then use that to get the actual pointer klass
        # then set the result to that pointer klass

        attr_path_and_name = f"syft.core.tensor.tensor.Tensor.__{op_str}__"

        result = TensorPointer(client=self.client)

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=[other], kwargs={})

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[other],
            kwargs={},
        )

        result_public_shape = None
        result_public_dtype = None

        if isinstance(other, TensorPointer):
            other_shape = other.public_shape
            other_dtype = other.public_dtype
        elif isinstance(other, (int, float)):
            other_shape = (1,)
            other_dtype = np.int32
        elif isinstance(other, bool):
            other_shape = (1,)
            other_dtype = np.dtype("bool")
        elif isinstance(other, np.ndarray):
            other_shape = other.shape
            other_dtype = other.dtype

        else:
            raise ValueError(f"Invalid Type for TensorPointer:{type(other)}")

        if self.public_shape is not None and other_shape is not None:
            result_public_shape = utils.get_shape(
                op_str, self.public_shape, other_shape
            )

        if self.public_dtype is not None and other_dtype is not None:
            if self.public_dtype != other_dtype:
                raise ValueError(
                    f"Type for self and other do not match ({self.public_dtype} vs {other_dtype})"
                )
            result_public_dtype = self.public_dtype

        result.public_shape = result_public_shape
        result.public_dtype = result_public_dtype

        return result

    @staticmethod
    def _apply_op(
        self: TensorPointer,
        other: Union[TensorPointer, MPCTensor, int, float, np.ndarray],
        op_str: str,
    ) -> Union[MPCTensor, TensorPointer]:
        """Performs the operation based on op_str

        Args:
            other (Union[TensorPointer,MPCTensor,int,float,np.ndarray]): second operand.

        Returns:
            Tuple[MPCTensor,Union[MPCTensor,int,float,np.ndarray]] : Result of the operation
        """
        op = getattr(operator, op_str)

        if isinstance(other, TensorPointer) and self.client != other.client:

            parties = [self.client, other.client]
            self_mpc = MPCTensor(secret=self, shape=self.public_shape, parties=parties)
            other_mpc = MPCTensor(
                secret=other, shape=other.public_shape, parties=parties
            )
            return op(self_mpc, other_mpc)

        elif isinstance(other, MPCTensor):

            return op(other, self)

        return self._apply_tensor_op(other=other, op_str=op_str)

    def __add__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "add" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorPointer._apply_op(self, other, "add")

    def __sub__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "sub" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorPointer._apply_op(self, other, "sub")

    def __mul__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "mul" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorPointer._apply_op(self, other, "mul")

    def __lt__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "lt" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "lt")

    def __gt__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "gt" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "gt")

    def __ge__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "ge" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "ge")

    def __le__(
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "le" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "le")

    def __eq__(  # type: ignore
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "eq" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "eq")

    def __ne__(  # type: ignore
        self, other: Union[TensorPointer, MPCTensor, int, float, np.ndarray]
    ) -> Union[TensorPointer, MPCTensor]:
        """Apply the "ne" operation between "self" and "other"

        Args:
            y (Union[TensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorPointer,MPCTensor] : Result of the operation.
        """

        return TensorPointer._apply_op(self, other, "ne")


def to32bit(np_array: np.ndarray, verbose: bool = True) -> np.ndarray:

    if np_array.dtype == np.int64:
        if verbose:
            print("Casting internal tensor to int32")
        out = np_array.astype(np.int32)

    elif np_array.dtype == np.float64:

        if verbose:
            print("Casting internal tensor to float32")
        out = np_array.astype(np.float32)

    else:
        out = np_array

    return out


@serializable(capnp_bytes=True)
class Tensor(
    PassthroughTensor,
    AutogradTensorAncestor,
    PhiTensorAncestor,
    FixedPrecisionTensorAncestor,
    # MPCTensorAncestor,
):

    # __attr_allowlist__ = ["child", "tag_name", "public_shape", "public_dtype"]

    PointerClassOverride = TensorPointer

    def __init__(
        self,
        child: Any,
        public_shape: Optional[Tuple[int, ...]] = None,
        public_dtype: Optional[np.dtype] = None,
    ) -> None:
        """data must be a list of numpy array"""

        if isinstance(child, (list, np.int32)):
            child = to32bit(np.array(child), verbose=False)

        if isinstance(child, th.Tensor):
            print(
                "Converting PyTorch tensor to numpy tensor for internal representation..."
            )
            child = to32bit(child.numpy())

        if not isinstance(child, PassthroughTensor) and not isinstance(
            child, np.ndarray
        ):

            raise Exception(
                f"Data: {child} ,type: {type(child)} must be list or nd.array "
            )

        if not isinstance(child, (np.ndarray, PassthroughTensor)) or (
            getattr(child, "dtype", None) not in [np.int32, np.bool_]
            and getattr(child, "dtype", None) is not None
        ):
            raise TypeError(
                "You tried to pass an a tensor of type:"
                + str(type(child))
                + " with child.dtype == "
                + str(getattr(child, "dtype", None))
                + ". Syft tensor objects only support np.int32 objects at this time. Please pass in either "
                "a list of int objects or a np.int32 array. We apologise for the inconvenience and will "
                "be adding support for more types very soon!"
            )

        kwargs = {"child": child}
        super().__init__(**kwargs)

        # set public shape to be the shape of the data since we have access to it at present
        if public_shape is None:
            public_shape = tuple(self.shape)

        # set public dtype to be the dtype of the data since we have access to it at present
        if public_dtype is None:
            public_dtype = str(self.dtype)

        self.tag_name: Optional[str] = None
        self.public_shape = public_shape
        self.public_dtype = public_dtype

    def tag(self, name: str) -> Tensor:
        self.tag_name = name
        return self

    def init_pointer(
        self,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
    ) -> Pointer:
        # relative
        from .autodp.single_entity_phi import SingleEntityPhiTensor
        from .autodp.single_entity_phi import TensorWrappedSingleEntityPhiTensorPointer

        # TODO:  Should create init pointer for NDimEntityPhiTensorPointer.

        if isinstance(self.child, SingleEntityPhiTensor):
            return TensorWrappedSingleEntityPhiTensorPointer(
                entity=self.child.entity,
                client=client,
                id_at_location=id_at_location,
                object_type=object_type,
                tags=tags,
                description=description,
                min_vals=self.child.min_vals,
                max_vals=self.child.max_vals,
                scalar_manager=self.child.scalar_manager,
                public_shape=getattr(self, "public_shape", None),
                public_dtype=getattr(self, "public_dtype", None),
            )
        else:
            return TensorPointer(
                client=client,
                id_at_location=id_at_location,
                object_type=object_type,
                tags=tags,
                description=description,
                public_shape=getattr(self, "public_shape", None),
                public_dtype=getattr(self, "public_dtype", None),
            )

    def bit_decomposition(self) -> Tensor:
        raise ValueError("Should not reach this point")

    def mpc_swap(self, other: Tensor) -> Tensor:
        self.child.child = other.child.child
        return self

    def _object2bytes(self) -> bytes:
        schema = get_capnp_schema(schema_file="tensor.capnp")
        tensor_struct: CapnpModule = schema.Tensor  # type: ignore
        tensor_msg = tensor_struct.new_message()

        # this is how we dispatch correct deserialization of bytes
        tensor_msg.magicHeader = serde_magic_header(type(self))

        chunk_bytes(sy.serialize(self.child, to_bytes=True), "child", tensor_msg)
        chunk_bytes(
            sy.serialize(self.public_shape, to_bytes=True), "publicShape", tensor_msg
        )
        chunk_bytes(
            sy.serialize(self.public_dtype, to_bytes=True), "publicDtype", tensor_msg
        )
        chunk_bytes(sy.serialize(self.tag_name, to_bytes=True), "tagName", tensor_msg)

        return tensor_msg.to_bytes_packed()

    @staticmethod
    def _bytes2object(buf: bytes) -> Tensor:
        schema = get_capnp_schema(schema_file="tensor.capnp")
        tensor_struct: CapnpModule = schema.Tensor  # type: ignore
        # https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize
        MAX_TRAVERSAL_LIMIT = 2**64 - 1
        tensor_msg = tensor_struct.from_bytes_packed(
            buf, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
        )

        tensor = Tensor(
            child=sy.deserialize(combine_bytes(tensor_msg.child), from_bytes=True),
            public_shape=sy.deserialize(
                combine_bytes(tensor_msg.publicShape), from_bytes=True
            ),
            public_dtype=sy.deserialize(
                combine_bytes(tensor_msg.publicDtype), from_bytes=True
            ),
        )
        tensor.tag_name = sy.deserialize(
            combine_bytes(tensor_msg.tagName), from_bytes=True
        )

        return tensor



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/python/bytes.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cproto/lib/python/bytes.proto\x12\x0fsyft.lib.python"\x15\n\x05\x42ytes\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x62\x06proto3'
)


_BYTES = DESCRIPTOR.message_types_by_name["Bytes"]
Bytes = _reflection.GeneratedProtocolMessageType(
    "Bytes",
    (_message.Message,),
    {
        "DESCRIPTOR": _BYTES,
        "__module__": "proto.lib.python.bytes_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.python.Bytes)
    },
)
_sym_db.RegisterMessage(Bytes)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _BYTES._serialized_start = 49
    _BYTES._serialized_end = 70
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from ......proto.core.node.common.service.resolve_pointer_type_service_pb2 import (
    ResolvePointerTypeAnswerMessage as ResolvePointerTypeAnswerMessage_PB,
)
from ......proto.core.node.common.service.resolve_pointer_type_service_pb2 import (
    ResolvePointerTypeMessage as ResolvePointerTypeMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@serializable()
class ResolvePointerTypeMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        id_at_location: UID,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.id_at_location = id_at_location

    def _object2proto(self) -> ResolvePointerTypeMessage_PB:
        return ResolvePointerTypeMessage_PB(
            id_at_location=sy.serialize(self.id_at_location),
            address=sy.serialize(self.address),
            msg_id=sy.serialize(self.id),
            reply_to=sy.serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: ResolvePointerTypeMessage_PB,
    ) -> "ResolvePointerTypeMessage":
        return ResolvePointerTypeMessage(
            id_at_location=sy.deserialize(blob=proto.id_at_location),
            address=sy.deserialize(blob=proto.address),
            msg_id=sy.deserialize(blob=proto.msg_id),
            reply_to=sy.deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return ResolvePointerTypeMessage_PB


@serializable()
class ResolvePointerTypeAnswerMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        type_path: str,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.type_path = type_path

    def _object2proto(self) -> ResolvePointerTypeAnswerMessage_PB:
        return ResolvePointerTypeAnswerMessage_PB(
            msg_id=sy.serialize(self.id),
            address=sy.serialize(self.address),
            type_path=self.type_path,
        )

    @staticmethod
    def _proto2object(
        proto: ResolvePointerTypeAnswerMessage_PB,
    ) -> "ResolvePointerTypeAnswerMessage":
        return ResolvePointerTypeAnswerMessage(
            msg_id=sy.deserialize(blob=proto.msg_id),
            address=sy.deserialize(blob=proto.address),
            type_path=proto.type_path,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return ResolvePointerTypeAnswerMessage_PB



"""ABY3 Protocol.

ABY3 : A Mixed Protocol Framework for Machine Learning.
https://eprint.iacr.org/2018/403.pdf
"""
# stdlib
from functools import reduce
import secrets
from typing import Any
from typing import List
from typing import Union
from uuid import UUID

# third party
import numpy as np
from tqdm import tqdm

# relative
from .....ast.klass import get_run_class_method
from ....common import UID
from ....tensor.smpc.mpc_tensor import MPCTensor
from ....tensor.smpc.share_tensor import ShareTensor
from ....tensor.smpc.utils import get_nr_bits
from ....tensor.tensor import Tensor
from ...store.crypto_primitive_provider import CryptoPrimitiveProvider


class ABY3:
    """ABY3 Protocol Implementation."""

    def __eq__(self, other: Any) -> bool:
        """Check if "self" is equal with another object given a set of attributes to compare.

        Args:
            other (Any): Object to compare

        Returns:
            bool: True if equal False if not.
        """
        if type(self) != type(other):
            return False

        return True

    @staticmethod
    def bit_injection(x: MPCTensor, ring_size: int) -> MPCTensor:
        """Perform ABY3 bit injection for conversion of binary share to arithmetic share.

        Args:
            x (MPCTensor) : MPCTensor with shares of bit.
            ring_size (int) : Ring size of arithmetic share to convert.

        Returns:
            arith_share (MPCTensor): Arithmetic shares of bit in input ring size.

        Raises:
            ValueError: If input tensor is not binary shared.
            ValueError: If the exactly three parties are not involved in the computation.
        """
        # relative
        # relative
        from ....tensor import TensorPointer

        shape = x.shape
        parties = x.parties
        seed_id_locations = secrets.randbits(64)
        kwargs = {"seed_id_locations": seed_id_locations}
        path_and_name = x.child[0].path_and_name
        attr_path_and_name = f"{x.child[0].path_and_name}.bit_decomposition"

        if not isinstance(x.child[0], TensorPointer):
            decomposed_shares = [
                share.bit_decomposition(share, ring_size, False, **kwargs)
                for share in x.child
            ]
        else:
            decomposed_shares = []
            op = get_run_class_method(attr_path_and_name, SMPC=True)
            for share in x.child:
                decomposed_shares.append(op(share, share, ring_size, False, **kwargs))

        decomposed_shares = ABY3.pregenerate_pointers(
            parties, 1, path_and_name, seed_id_locations
        )

        # List which contains the share of a single bit
        res_shares: List[MPCTensor] = []

        bit_shares = [share[0] for share in decomposed_shares]
        bit_shares = zip(*bit_shares)  # type: ignore
        for bit_sh in bit_shares:
            mpc = MPCTensor(
                shares=bit_sh, shape=shape, parties=parties, ring_size=ring_size
            )
            res_shares.append(mpc)

        # TODO: Should modify to xor at mpc tensor level
        arith_share = reduce(lambda a, b: a + b - (2 * a * b), res_shares)

        return arith_share

    @staticmethod
    def full_adder(a: List[MPCTensor], b: List[MPCTensor]) -> List[MPCTensor]:
        """Perform bit addition on MPCTensors using a full adder.

        Args:
            a (List[MPCTensor]): MPCTensor with shares of bit.
            b (List[MPCTensor]): MPCTensor with shares of bit.

        Returns:
            result (List[MPCTensor]): Result of the operation.

        TODO: Should modify ripple carry adder to parallel prefix adder.
        """
        parties = a[0].parties
        parties_info = a[0].parties_info

        shape_x = tuple(a[0].shape)  # type: ignore
        shape_y = tuple(b[0].shape)  # type: ignore
        ring_size = 2**32

        # For ring_size 2 we generate those before hand
        CryptoPrimitiveProvider.generate_primitives(
            "beaver_mul",
            nr_instances=64,
            parties=parties,
            g_kwargs={
                "a_shape": shape_x,
                "b_shape": shape_y,
                "parties_info": parties_info,
            },
            p_kwargs={"a_shape": shape_x, "b_shape": shape_y},
            ring_size=2,
        )

        ring_bits = get_nr_bits(ring_size)
        c = np.array([0], dtype=np.bool)  # carry bits of addition.
        result: List[MPCTensor] = []
        for idx in range(ring_bits):
            print("Bit Number :", idx)
            s_tmp = a[idx] + b[idx]
            s = s_tmp + c
            if idx != ring_bits - 1:
                c = a[idx] * b[idx] + c * s_tmp
                c.block
            result.append(s)
        return result

    @staticmethod
    def full_adder_spdz_compiler(
        a: List[MPCTensor], b: List[MPCTensor]
    ) -> List[MPCTensor]:
        # Specialized for 3 parties
        """Perform bit addition on MPCTensors using a full adder.

        Args:
            a (List[MPCTensor]): MPCTensor with shares of bit.
            b (List[MPCTensor]): MPCTensor with shares of bit.

        Returns:
            result (List[MPCTensor]): Result of the operation.

        TODO: Should modify ripple carry adder to parallel prefix adder.
        """
        parties = a[0].parties
        parties_info = a[0].parties_info

        shape_x = tuple(a[0].shape)  # type: ignore
        shape_y = tuple(b[0].shape)  # type: ignore
        ring_size = 2**32

        # For ring_size 2 we generate those before hand
        CryptoPrimitiveProvider.generate_primitives(
            "beaver_mul",
            nr_instances=32,
            parties=parties,
            g_kwargs={
                "a_shape": shape_x,
                "b_shape": shape_y,
                "parties_info": parties_info,
            },
            p_kwargs={"a_shape": shape_x, "b_shape": shape_y},
            ring_size=2,
        )

        ring_bits = get_nr_bits(ring_size)

        carry = np.zeros(a[0].mpc_shape, dtype=np.bool)
        one = MPCTensor(
            parties=parties,
            secret=Tensor(np.ones(a[0].mpc_shape, dtype=np.bool)),
            shape=a[0].mpc_shape,
        )

        result: List[MPCTensor] = []

        def majority(
            a: Union[MPCTensor, np.ndarray],
            b: Union[MPCTensor, np.ndarray],
            c: Union[MPCTensor, np.ndarray],
        ) -> MPCTensor:
            return (a + c + one) * (b + c) + b

        for idx in tqdm(range(ring_bits), desc="Computing..."):
            s = a[idx] + b[idx] + carry
            if idx != ring_bits - 1:
                carry = majority(a[idx], b[idx], carry)
                # time.sleep(1)
                carry.block
            result.append(s)
        return result

    @staticmethod
    def bit_decomposition(x: MPCTensor) -> List[MPCTensor]:
        """Perform ABY3 bit decomposition for conversion of arithmetic share to binary share.

        Args:
            x (MPCTensor): Arithmetic shares of secret.

        Returns:
            bin_share (List[MPCTensor]): Returns binary shares of each bit of the secret.

        TODO : Should be modified to use parallel prefix adder when multiprocessing
        functionality is integrated
        """
        # relative
        from ....tensor import TensorPointer

        nr_parties = len(x.parties)
        ring_size = 2**32  # Should extract this info better
        ring_bits = get_nr_bits(ring_size)
        shape = x.shape
        parties = x.parties

        seed_id_locations = secrets.randbits(64)
        kwargs = {"seed_id_locations": seed_id_locations}
        path_and_name = x.child[0].path_and_name
        attr_path_and_name = f"{x.child[0].path_and_name}.bit_decomposition"

        if not isinstance(x.child[0], TensorPointer):
            decomposed_shares = [
                share.bit_decomposition(share, 2, True, **kwargs) for share in x.child
            ]
        else:
            decomposed_shares = []
            op = get_run_class_method(attr_path_and_name, SMPC=True)
            for share in x.child:
                decomposed_shares.append(op(share, share, 2, True, **kwargs))

        decomposed_shares = ABY3.pregenerate_pointers(
            parties, ring_bits, path_and_name, seed_id_locations
        )

        # List which contains the share of each share.
        # TODO: Shouldn't this be an empty list? and we append to it?
        res_shares: List[List[MPCTensor]] = [[] for _ in range(nr_parties)]

        for idx in range(ring_bits):
            bit_shares = [share[idx] for share in decomposed_shares]
            bit_shares = zip(*bit_shares)  # type: ignore
            for i, bit_sh in enumerate(bit_shares):
                mpc = MPCTensor(
                    shares=bit_sh, shape=shape, parties=parties, ring_size=2
                )
                res_shares[i].append(mpc)

        if nr_parties == 2:
            # Specialized for two parties
            bin_share = ABY3.full_adder_spdz_compiler(res_shares[0], res_shares[1])
        else:
            bin_share = reduce(ABY3.full_adder, res_shares)

        return bin_share

    @staticmethod
    def pregenerate_pointers(
        parties: List[Any], ring_bits: int, path_and_name: str, seed_id_locations: int
    ) -> List[List[List[ShareTensor]]]:
        generator = np.random.default_rng(seed_id_locations)
        # Skip the first ID ,as it is used for None return type in run class method.
        _ = UID(UUID(bytes=generator.bytes(16)))

        nr_parties = len(parties)
        resolved_pointer_type = [
            party.lib_ast.query(path_and_name) for party in parties
        ]

        """
        Consider bit decomposition.
        We create a share of share such that.
        Assume we are operating in ring_size 2**32(32 bits)
        Since we operate in n-out-of-n secret sharing, each party has a single share.
        Consider two parties such that a secret x is split as
        x = x1+x2
        To create share of shares,we use the intuition that the shares not held by party are made zero

                    Party1        Party2

        x1           x1             0

        x2           0               x2

        x_i_j denotes the shares held by jth party of ith share

        x_1_1 = x1
        x_1_2 = 0
        x_2_1 = 0
        x_2_2 = x2

        Party 1 = [x_1_1,x_2_1]
        Party 2  =[x_1_2,x_2_2]

        Now each party party has share of shares
        In bit decomposition, we split each bit and create share of shares

        Party 1 = [ [share of shares of first bit] ,[...second bit] ...[ nth bit]]

        Note: Count (share of shares for a particular bit) = number of parties
        """
        share_pointers: List[List[List[Any]]] = [[] for _ in range(nr_parties)]

        for _ in range(ring_bits * nr_parties):
            id_at_location = UID(UUID(bytes=generator.bytes(16)))
            for idx, party in enumerate(parties):
                result = resolved_pointer_type[idx].pointer_type(client=party)
                result.id_at_location = id_at_location
                share_pointers[idx].append(result)

        share_pointers = [
            [
                share_lst[i : i + nr_parties]  # noqa
                for i in range(0, len(share_lst), nr_parties)
            ]
            for share_lst in share_pointers
        ]

        return share_pointers



# CLEANUP NOTES:
# - remove unused comments
# - add documentation for each method
# - add comments inline explaining each piece
# - add a unit test for each method (at least)

# stdlib
from functools import lru_cache
import math
from typing import Dict as TypeDict
from typing import Iterable
from typing import KeysView as TypeKeysView
from typing import List as TypeList
from typing import Optional
from typing import Set as TypeSet
from typing import Tuple
from typing import Union

# third party
from autodp.autodp_core import Mechanism
from autodp.transformer_zoo import Composition
from nacl.signing import VerifyKey
from sqlalchemy.engine import Engine

# relative
from ..node.common.node_manager.ledger_manager import AbstractLedger
from ..node.common.node_manager.ledger_manager import DatabaseLedger
from ..node.common.node_manager.ledger_manager import DictLedger
from .entity import DataSubjectGroup
from .entity import Entity
from .idp_gaussian_mechanism import iDPGaussianMechanism


def compose_mechanisms(
    mechanisms: Iterable[iDPGaussianMechanism], delta: float
) -> float:
    sigmas = list()
    squared_l2_norms = list()
    squared_l2_norm_upper_bounds = list()
    Ls = list()
    values = list()

    for m in mechanisms:
        sigmas.append(m.params["sigma"])
        squared_l2_norms.append(m.params["private_value"])
        squared_l2_norm_upper_bounds.append(m.params["public_value"])
        Ls.append(m.params["L"])
        values.append(m.params["value"])

    return compose_mechanisms_via_simplified_args_for_lru_cache(
        tuple(sigmas),
        tuple(squared_l2_norms),
        tuple(squared_l2_norm_upper_bounds),
        tuple(Ls),
        tuple(values),
        delta,
    )


@lru_cache(maxsize=None)
def compose_mechanisms_via_simplified_args_for_lru_cache(
    sigmas: Tuple[float],
    squared_l2_norms: Tuple[float],
    squared_l2_norm_upper_bounds: Tuple[float],
    Ls: Tuple[float],
    values: Tuple[float],
    delta: float,
) -> float:
    mechanisms = list()
    for i in range(len(sigmas)):

        m = iDPGaussianMechanism(
            sigma=sigmas[i],
            squared_l2_norm=squared_l2_norms[i],
            squared_l2_norm_upper_bound=squared_l2_norm_upper_bounds[i],
            L=Ls[i],
            entity_name="",
        )
        m.params["value"] = values[i]
        mechanisms.append(m)
    # compose them with the transformation: compose
    compose = Composition()
    composed_mech = compose(mechanisms, [1] * len(mechanisms))
    eps = composed_mech.get_approxDP(delta)
    return eps


class AdversarialAccountant:
    """Adversarial Accountant class that keeps track of budget and maintains a privacy ledger."""

    def __init__(
        self, db_engine: Engine = None, max_budget: float = 10, delta: float = 1e-6
    ) -> None:

        if db_engine is not None:
            # this is a database-backed lookup table
            # maps an entity to an actual budget
            self.entity2ledger: AbstractLedger = DatabaseLedger(db_engine)
        else:
            self.entity2ledger: AbstractLedger = DictLedger()  # type: ignore

        # this is a temporary lookup table for mechanisms we're not sure
        # we're going to keep (See publish.py for how this is used)
        self.temp_entity2ledger: TypeDict = {}
        self.max_budget = max_budget
        self.delta = delta

    def temp_append(
        self, entity2mechanisms: TypeDict[Entity, TypeList[Mechanism]]
    ) -> None:
        # add the mechanisms to all of the entities
        for key, ms in entity2mechanisms.items():
            if key not in self.temp_entity2ledger.keys():
                self.temp_entity2ledger[key] = list()
            for m in ms:
                self.temp_entity2ledger[key].append(m)

    def append(self, entity2mechanisms: TypeDict[str, TypeList[Mechanism]]) -> None:
        mechanisms = list()
        # add all the mechanisms
        for _, ms in entity2mechanisms.items():
            for m in ms:
                mechanisms.append(m)

        self.entity2ledger.register_mechanisms(mechanisms)

    # save the temporary ledger into the database
    def save_temp_ledger_to_longterm_ledger(self) -> None:
        self.append(entity2mechanisms=self.temp_entity2ledger)

    # return epsilons for each entity
    def get_eps_for_entity(
        self,
        entity: Entity,
        user_key: Optional[VerifyKey] = None,
        returned_epsilon_is_private: bool = False,
    ) -> float:

        # fetch mechanisms from the database
        table_mechanisms = self.entity2ledger.query(entity_name=entity.name)  # type: ignore
        mechanisms = [x.obj for x in table_mechanisms]

        # filter out mechanisms that weren't created by this data scientist user
        if user_key is not None:
            filtered_mechanisms = []
            for mech in mechanisms:
                if mech.user_key == user_key:
                    filtered_mechanisms.append(mech)

            mechanisms = filtered_mechanisms

        if entity in self.temp_entity2ledger.keys():
            mechanisms = mechanisms + self.temp_entity2ledger[entity]

        # # filter out mechanisms that weren't created by this data scientist user
        # if user_key is not None:
        #     filtered_mechanisms = []
        #     for mech in mechanisms:
        #         # left = _deserialize(mech.user_key, from_bytes=True)
        #         # print("Comparing Left:" + str(mech.user_key) + " of type " + str(type(mech.user_key)))
        #         # print("Comparing Right:" + str(user_key) + " of type " + str(type(user_key)))

        #         # user_key = VerifyKey(user_key.encode("utf-8"), encoder=HexEncoder)

        #         if mech.user_key == user_key:
        #             filtered_mechanisms.append(mech)

        #     mechanisms = filtered_mechanisms

        # print("Num mechanisms after TEMP:" + str(len(mechanisms)))
        # for m in mechanisms:
        # print("Filtered Mechanism Entity:" + str(m.entity_name))

        # print("Mechanisma after filtering: ", mechanisms)
        # use verify key to specify the user
        # for all entities in the db,
        # how do we ensure that no data scientist
        # exceeds the budget of any entity?

        if returned_epsilon_is_private:
            for mech in mechanisms:
                mech.params["value"] = mech.params["private_value"]
        else:
            for mech in mechanisms:
                mech.params["value"] = mech.params["public_value"]

        # map dataset
        if len(mechanisms) > 0:
            eps = compose_mechanisms(tuple(mechanisms), self.delta)

            # if we have a user key, get the budget and clamp the epsilon
            # if user_key is not None:
            #     user_budget = self.entity2ledger.get_user_budget(user_key=user_key)
            #     print("USER BUDGET:" + str(user_budget))
            #     print(f"EPS: {eps}")
            #     if eps > user_budget:
            #         print("setting the eps to the user budget")
            #         eps = user_budget
        else:
            eps = 0

        return float(eps)

        # # Query for eps given delta
        # return PhiScalar(
        #     value=composed_mech.get_approxDP(self.delta),
        #     min_val=0,
        #     max_val=self.max_budget,
        #     entity=entity,
        # )

    # checks if the entity has budget or not
    def has_budget(
        self,
        entity: Entity,
        user_key: VerifyKey,
        returned_epsilon_is_private: bool = False,
    ) -> bool:

        spent = self.get_eps_for_entity(
            entity=entity,
            user_key=user_key,
            returned_epsilon_is_private=returned_epsilon_is_private,
        )

        user_budget = self.entity2ledger.get_user_budget(user_key=user_key)

        # print("ACCOUNTANT MAX BUDGET", self.max_budget)
        # @Andrew can we use <= or does it have to be <
        has_budget = spent <= user_budget
        # print("\n\nHas Budget:" + str(has_budget))
        # print("YOU'VE SPENT:" + str(spent))
        # print("USER BUDGET:" + str(user_budget))
        return has_budget

    # returns maximum entity epsilon
    def user_budget(
        self, user_key: VerifyKey, returned_epsilon_is_private: bool = False
    ) -> float:
        max_spend = 0.0

        for ent in self.entities:
            spend = self.get_eps_for_entity(
                entity=ent,
                user_key=user_key,
                returned_epsilon_is_private=returned_epsilon_is_private,
            )
            if math.isnan(spend) or math.isinf(spend):
                print(f"Warning: Spend is {spend}")

            if spend > max_spend:
                max_spend = float(spend)

        return float(max_spend)

    def get_remaining_budget(
        self, user_key: VerifyKey, returned_epsilon_is_private: bool = False
    ) -> float:
        max_spend = self.user_budget(
            user_key=user_key, returned_epsilon_is_private=returned_epsilon_is_private
        )
        if math.isnan(max_spend) or math.isinf(max_spend):
            print(f"Warning: Remaining budget not valid with max_spend {max_spend}")

        # If there's no entity registered
        if not len(self.entities):
            max_spend = 0

        user_budget = self.entity2ledger.get_user_budget(user_key=user_key)

        return float(user_budget - max_spend)

    @property
    def entities(self) -> TypeKeysView[Entity]:
        return self.entity2ledger.keys()

    # returns a collection of entities having no budget
    def overbudgeted_entities(
        self,
        temp_entities: Union[
            TypeDict[Entity, TypeList[iDPGaussianMechanism]],
            TypeDict[DataSubjectGroup, TypeList[iDPGaussianMechanism]],
        ],
        user_key: VerifyKey,
        returned_epsilon_is_private: bool = False,
    ) -> Union[TypeSet[Entity], TypeSet[DataSubjectGroup]]:
        entities = set()

        for entity, _ in temp_entities.items():
            if isinstance(entity, DataSubjectGroup):
                for e in entity.entity_set:
                    if not self.has_budget(
                        e,
                        user_key=user_key,
                        returned_epsilon_is_private=returned_epsilon_is_private,
                    ):
                        entities.add(
                            entity
                        )  # Leave out the whole group if ANY of its entities are over budget
            elif isinstance(entity, Entity):
                if not self.has_budget(
                    entity,
                    user_key=user_key,
                    returned_epsilon_is_private=returned_epsilon_is_private,
                ):
                    entities.add(entity)  # type: ignore
            else:
                raise Exception
        return entities

    # prints entity and its epsilon value
    def print_ledger(self, returned_epsilon_is_private: bool = False) -> None:
        for entity in self.entities:
            if entity is not None:
                print(
                    str(entity.name)
                    + "\t"
                    + str(
                        self.get_eps_for_entity(
                            entity=entity,
                            returned_epsilon_is_private=returned_epsilon_is_private,
                        )
                    )
                )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/common/common_object.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n%proto/core/common/common_object.proto\x12\x10syft.core.common"x\n\x0c\x43ommonObject\x12)\n\nuid_object\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12=\n\x15object_with_id_object\x18\x02 \x01(\x0b\x32\x1e.syft.core.common.ObjectWithID"1\n\x0cObjectWithID\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID"\x14\n\x03UID\x12\r\n\x05value\x18\x01 \x01(\x0c\x62\x06proto3'
)


_COMMONOBJECT = DESCRIPTOR.message_types_by_name["CommonObject"]
_OBJECTWITHID = DESCRIPTOR.message_types_by_name["ObjectWithID"]
_UID = DESCRIPTOR.message_types_by_name["UID"]
CommonObject = _reflection.GeneratedProtocolMessageType(
    "CommonObject",
    (_message.Message,),
    {
        "DESCRIPTOR": _COMMONOBJECT,
        "__module__": "proto.core.common.common_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.common.CommonObject)
    },
)
_sym_db.RegisterMessage(CommonObject)

ObjectWithID = _reflection.GeneratedProtocolMessageType(
    "ObjectWithID",
    (_message.Message,),
    {
        "DESCRIPTOR": _OBJECTWITHID,
        "__module__": "proto.core.common.common_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.common.ObjectWithID)
    },
)
_sym_db.RegisterMessage(ObjectWithID)

UID = _reflection.GeneratedProtocolMessageType(
    "UID",
    (_message.Message,),
    {
        "DESCRIPTOR": _UID,
        "__module__": "proto.core.common.common_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.common.UID)
    },
)
_sym_db.RegisterMessage(UID)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _COMMONOBJECT._serialized_start = 59
    _COMMONOBJECT._serialized_end = 179
    _OBJECTWITHID._serialized_start = 181
    _OBJECTWITHID._serialized_end = 230
    _UID._serialized_start = 232
    _UID._serialized_end = 252
# @@protoc_insertion_point(module_scope)



# CLEANUP NOTES:
# - add documentation for each method
# - add comments inline explaining each piece
# - add a unit test for each method (at least)

# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
import names

# relative
from ...proto.core.adp.entity_pb2 import DataSubjectGroup as DataSubjectGroup_PB
from ...proto.core.adp.entity_pb2 import Entity as Entity_PB
from ..common import UID
from ..common.serde.deserialize import _deserialize as deserialize
from ..common.serde.serializable import serializable
from ..common.serde.serialize import _serialize as serialize


@serializable()
class Entity:
    __slots__ = "name"

    def __init__(self, name: str = "") -> None:

        # If someone doesn't provide a unique name - make one up!
        if name == "":
            name = names.get_full_name().replace(" ", "_") + "_g"

        if ";" in name:
            raise Exception(
                "Entity name cannot contain a semi-colon (;), sorry. Don't ask why. Accept your fate."
            )

        if "+" in name:
            raise Exception(
                "Entity name cannot contain a plus (+), sorry. Don't ask why. Accept your fate."
            )

        self.name = name

    @property
    def attributes(self) -> Dict[str, str]:
        return {"name": self.name}

    # returns a hash value for the entity
    def __hash__(self) -> int:
        return hash(self.name)

    # checks if the two entities are equal
    def __eq__(self, other: Any) -> bool:
        # TODO: Remove this once Entity is refactored out
        if isinstance(other, str):
            return self.name == other
        return self.name == other.name

    # checks if the two entities are not equal
    def __ne__(self, other: Any) -> bool:
        # Not strictly necessary, but to avoid having both x==y and x!=y
        # True at the same time
        return hash(self) != hash(other)

    def __lt__(self, other: Any) -> bool:
        return self.name < other.name

    def __add__(
        self, other: Union[Entity, DataSubjectGroup, int, float]
    ) -> Union[DataSubjectGroup, Entity]:
        if isinstance(other, Entity):
            return DataSubjectGroup([self, other])
        elif isinstance(other, DataSubjectGroup):
            other.entity_set.add(self)
            return other
        elif not other:  # type: ignore
            return DataSubjectGroup([self])
        elif isinstance(other, (int, float)):
            return self
        else:
            raise Exception(
                f"Addition not implemented between {type(self)} and {type(other)}"
            )

    def __mul__(
        self, other: Union[Entity, DataSubjectGroup, int, float]
    ) -> Union[DataSubjectGroup, Entity]:
        return self.__add__(other)

    def to_string(self) -> str:
        return f"{self.name}"

    @staticmethod
    def from_string(blob: str) -> Entity:
        return Entity(name=blob)

    # represents entity as a string
    def __repr__(self) -> str:
        return "<Entity:" + str(self.name) + ">"

    # converts entity into a protobuf object
    def _object2proto(self) -> Entity_PB:
        return Entity_PB(name=self.name)

    # converts a generated protobuf object into an entity
    @staticmethod
    def _proto2object(proto: Entity_PB) -> Entity:
        return Entity(name=proto.name)

    # returns the type of generated protobuf object
    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Entity_PB


@serializable()
class DataSubjectGroup:
    """Data Subject is what we have been calling an 'ENTITY' all along ..."""

    def __init__(
        self,
        list_of_entities: Optional[Union[list, set, Entity]] = None,
        id: Optional[UID] = None,
    ):
        self.entity_set: set = set()
        # Ensure each entity being tracked is unique
        if isinstance(list_of_entities, (list, set)):
            self.entity_set = self.entity_set.union(list_of_entities)
        elif isinstance(list_of_entities, Entity):
            self.entity_set.add(list_of_entities)  # type: ignore
        elif list_of_entities is None:  # Don't need to do anything if is NoneType
            pass
        else:
            raise Exception(
                f"Cannot initialize DSG with {type(list_of_entities)} - please try list or set instead."
            )
        self.id = id if id else UID()

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.entity_set)))

    def __eq__(self, other: DataSubjectGroup) -> bool:  # type: ignore
        return hash(self) == hash(other)

    def __contains__(self, item: Entity) -> bool:
        return item in self.entity_set

    def to_string(self) -> str:
        output_string = ";".join(item.to_string() for item in self.entity_set)
        return output_string

    @staticmethod
    def from_string(blob: str) -> DataSubjectGroup:
        """Take the output of to_string and recreate the DataSubjectGroup"""
        entity_list = blob.split(";")
        entity_set = set()
        for entity_blob in entity_list:
            entity_set.add(Entity.from_string(entity_blob))
        return DataSubjectGroup(entity_set)  # type: ignore

    def __add__(
        self, other: Union[DataSubjectGroup, Entity, int, float]
    ) -> DataSubjectGroup:
        if isinstance(other, Entity):
            return DataSubjectGroup(self.entity_set.union({other}))
        elif isinstance(other, DataSubjectGroup):
            return DataSubjectGroup(self.entity_set.union(other.entity_set))
        elif not other:  # type: ignore
            return self
        elif isinstance(other, (int, float)):
            return self
        else:
            raise Exception(
                f"Addition not implemented between {type(self)} and {type(other)}"
            )

    def __mul__(
        self, other: Union[DataSubjectGroup, Entity, int, float]
    ) -> DataSubjectGroup:
        return self.__add__(other)

    def __repr__(self) -> str:
        return f"DSG{[i.__repr__() for i in self.entity_set]}"

    # converts entity into a protobuf object
    def _object2proto(self) -> DataSubjectGroup_PB:
        return DataSubjectGroup_PB(
            id=self.id._object2proto(), entities=[serialize(x) for x in self.entity_set]
        )

    # converts a generated protobuf object into an entity
    @staticmethod
    def _proto2object(proto: DataSubjectGroup_PB) -> DataSubjectGroup:
        return DataSubjectGroup(
            list_of_entities=[deserialize(x) for x in proto.entities],
            id=UID._proto2object(proto.id),
        )

    # returns the type of generated protobuf object
    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return DataSubjectGroup_PB



# CLEANUP NOTES:
# - remove unused comments
# - add documentation for each method
# - add comments inline explaining each piece
# - add a unit test for each method (at least)

# stdlib
from copy import deepcopy
import random
from typing import Any
from typing import Dict as TypeDict
from typing import List
from typing import List as TypeList
from typing import Tuple
from typing import Union

# third party
from nacl.signing import VerifyKey
import numpy as np
from pymbolic.mapper.substitutor import SubstitutionMapper
from pymbolic.mapper.substitutor import make_subst_func

# relative
from .entity import DataSubjectGroup
from .entity import Entity
from .idp_gaussian_mechanism import iDPGaussianMechanism
from .search import max_lipschitz_wrt_entity


def publish(
    scalars: TypeList[Any],
    acc: Any,
    user_key: VerifyKey,
    sigma: float = 1.5,
    public_only: bool = False,
) -> TypeList[Any]:
    """
    Method to compute and update the mechanism values of data entities managed by the Adversarial Accountant
    class.
    """

    # Creates a temporary register (memory) to
    # store entities and their respective mechanisms
    acc.temp_entity2ledger = {}
    print("publish.py:45: TRY: ms = get_all_entity_mechanisms")
    # Recover all entity mechanisms
    ms = get_all_entity_mechanisms(
        scalars=scalars, sigma=sigma, public_only=public_only
    )
    print("publish.py:50: SUCCESS: ms = get_all_entity_mechanisms")

    print("publish.py:52: TRY: for _, mechs in ms.items():")
    # add the user_key to all of the mechanisms
    for _, mechs in ms.items():
        for m in mechs:
            m.user_key = user_key
    print("publish.py:57: SUCCESS: for _, mechs in ms.items():")

    print("publish.py:59: TRY: acc.temp_append(ms)")
    # Register the mechanism / entities at the temporary data structure
    # This data structure will be organized as a dictionary of
    # lists, each list will contain a set of mechanisms related to an entity.
    # Example: acc.temp_entity2ledger = {"patient_1": [<iDPGaussianMechanism>, <iDPGaussianMechanism>] }
    acc.temp_append(ms)
    print("publish.py:65: SUCCESS: acc.temp_append(ms)")

    print("publish.py:67: TRY: overbudgeted_entities = acc.overbudgeted_entities(")
    # Filter entities by searching for the overbudgeted ones.
    overbudgeted_entities = acc.overbudgeted_entities(
        temp_entities=acc.temp_entity2ledger,
        user_key=user_key,
        returned_epsilon_is_private=True,
    )
    print("publish.py:74: SUCCESS: overbudgeted_entities = acc.overbudgeted_entities(")

    print(
        "publish.py:76: TRY:  if len(overbudgeted_entities) > 0: scalars = deepcopy(scalars)"
    )
    # so that we don't modify the original polynomial
    # it might be fine to do so but just playing it safe
    if len(overbudgeted_entities) > 0:
        scalars = deepcopy(scalars)
    print(
        "publish.py:81: SUCCESS:  if len(overbudgeted_entities) > 0: scalars = deepcopy(scalars)"
    )

    # If some overbudgeted entity is found, run this.
    iterator = 0
    while len(overbudgeted_entities) > 0 and iterator < 3:
        print(
            "publish.py:86: INSIDE:  while len(overbudgeted_entities) > 0 and iterator < 3:"
        )
        iterator += 1
        print(
            "publish.py:88: len(overbudgeted_entities) == "
            + str(len(overbudgeted_entities))
        )
        print("publish.py:89:  for output_scalar in scalars:")
        input_scalars = set()
        for output_scalar in scalars:

            print("publish.py:93:  for input_scalar in output_scalar.input_scalars:")
            # output_scalar.input_scalars is a @property which determines
            # what inputs are still contributing to the output scalar
            # given that we may have just removed some
            for input_scalar in output_scalar.input_scalars:
                input_scalars.add(input_scalar)
        print("publish.py:99:  for input_scalar in input_scalars:")
        for input_scalar in input_scalars:
            if input_scalar.entity in overbudgeted_entities:
                for output_scalar in scalars:

                    # remove input_scalar from the computation that creates
                    # output scalar because this input_scalar is causing
                    # the budget spend to be too high.
                    output_scalar.poly = SubstitutionMapper(
                        make_subst_func({input_scalar.poly.name: 0})
                    )(output_scalar.poly)

        print("publish.py:110:  acc.temp_entity2ledger = {}")
        acc.temp_entity2ledger = {}

        print("publish.py:114:  BEGIN: ms = get_all_entity_mechanisms(")
        # get mechanisms for new publish event
        ms = get_all_entity_mechanisms(
            scalars=scalars, sigma=sigma, public_only=public_only
        )
        print("publish.py:119:  SUCCESS: ms = get_all_entity_mechanisms(")

        for _, mechs in ms.items():
            for m in mechs:
                m.user_key = user_key

        print("publish.py:127:  TRY: acc.temp_append(ms)")
        # this is when we actually insert into the database
        acc.temp_append(ms)
        print("publish.py:127:  SUCCESS: acc.temp_append(ms)")

        print(
            "publish.py:130:  TRY:  overbudgeted_entities = acc.overbudgeted_entities("
        )
        overbudgeted_entities = acc.overbudgeted_entities(
            temp_entities=acc.temp_entity2ledger,
            user_key=user_key,
            returned_epsilon_is_private=True,
        )
        print(
            "publish.py:136:  SUCCESS:  overbudgeted_entities = acc.overbudgeted_entities("
        )

    print(
        "publish.py:138:  TRY:  output = [s.value + random.gauss(0, sigma) for s in scalars]"
    )
    # Add to each scalar a a gaussian noise in an interval between
    # 0 to sigma value.
    output = [s.value + random.gauss(0, sigma) for s in scalars]
    print(
        "publish.py:142:  SUCCESS:  output = [s.value + random.gauss(0, sigma) for s in scalars]"
    )

    print("publish.py:144:  TRY:  acc.save_temp_ledger_to_longterm_ledger()")
    # Persist the temporary ledger into the database.
    acc.save_temp_ledger_to_longterm_ledger()
    print("publish.py:147:  SUCCESS:  acc.save_temp_ledger_to_longterm_ledger()")

    return output


def get_mechanism_for_entity(
    scalars: TypeList[Any],
    entity: Union[Entity, DataSubjectGroup],
    sigma: float = 1.5,
    public_only: bool = False,
) -> Union[
    List[Tuple[Entity, iDPGaussianMechanism]],
    List[Tuple[DataSubjectGroup, iDPGaussianMechanism]],
]:
    """
    Iterates over scalars computing its value and L attribute and builds its mechanism.
    """
    m_id = "ms_"
    for s in scalars:
        m_id += str(s.id).split(" ")[1][:-1] + "_"

    value_upper_bound = np.sqrt(
        np.sum(
            np.square(
                np.array([(float(s.max_val) - float(s.min_val)) for s in scalars])
            )
        )
    )

    if public_only:
        value = value_upper_bound
    else:
        value = np.sqrt(np.sum(np.square(np.array([float(s.value) for s in scalars]))))

    if isinstance(entity, DataSubjectGroup):
        mechanisms = []

        for e in entity.entity_set:
            mechanisms.append(
                (
                    entity,
                    iDPGaussianMechanism(
                        sigma=sigma,
                        squared_l2_norm=value,
                        squared_l2_norm_upper_bound=value_upper_bound,
                        L=float(max_lipschitz_wrt_entity(scalars, entity=e)),
                        entity_name=e.name,
                        name=m_id,
                    ),
                )
            )
        return mechanisms
    elif isinstance(entity, Entity):
        return [
            (
                entity,
                iDPGaussianMechanism(
                    sigma=sigma,
                    squared_l2_norm=value,
                    squared_l2_norm_upper_bound=value_upper_bound,
                    L=float(max_lipschitz_wrt_entity(scalars, entity=entity)),
                    entity_name=entity.name,
                    name=m_id,
                ),
            )
        ]
    else:
        raise Exception(f"What even is this type: {type(entity)}")


def get_all_entity_mechanisms(
    scalars: TypeList[Any], sigma: float = 1.5, public_only: bool = False
) -> TypeDict[Entity, Any]:
    "Generates a list of entities by processing the given scalars list."
    entities = set()
    for s in scalars:
        for i_s in s.input_scalars:
            entities.add(i_s.entity)
    entity_to_mechanisms: dict = {}
    for entity in entities:
        for flat_entity, mechanism in get_mechanism_for_entity(
            scalars=scalars, entity=entity, sigma=sigma, public_only=public_only
        ):
            # We ignore entity in this case because entity might be a DataSubjectGroup,
            # in which case flat_entity will include more entities than entity
            if flat_entity not in entity_to_mechanisms:
                entity_to_mechanisms[flat_entity] = list()
            entity_to_mechanisms[flat_entity].append(mechanism)
    return entity_to_mechanisms


def get_remaining_budget(acc: Any, user_key: VerifyKey) -> Any:
    budget = acc.get_remaining_budget(
        user_key=user_key, returned_epsilon_is_private=False
    )
    return budget



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/association_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n.proto/grid/messages/association_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xbb\x02\n\x1dSendAssociationRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0e\n\x06source\x18\x03 \x01(\t\x12\x0e\n\x06target\x18\x04 \x01(\t\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address\x12Q\n\x08metadata\x18\x06 \x03(\x0b\x32?.syft.grid.messages.SendAssociationRequestMessage.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xd3\x02\n ReceiveAssociationRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12T\n\x08metadata\x18\x03 \x03(\x0b\x32\x42.syft.grid.messages.ReceiveAssociationRequestMessage.MetadataEntry\x12\x10\n\x08response\x18\x04 \x01(\t\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0e\n\x06source\x18\x06 \x01(\t\x12\x0e\n\x06target\x18\x07 \x01(\t\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xcc\x01\n RespondAssociationRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x10\n\x08response\x18\x03 \x01(\t\x12\x0e\n\x06source\x18\x04 \x01(\t\x12\x0e\n\x06target\x18\x05 \x01(\t\x12\'\n\x08reply_to\x18\x06 \x01(\x0b\x32\x15.syft.core.io.Address"\xae\x01\n\x1cGetAssociationRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x16\n\x0e\x61ssociation_id\x18\x03 \x01(\x05\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8f\x02\n\x1dGetAssociationRequestResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12O\n\x07\x63ontent\x18\x02 \x03(\x0b\x32>.syft.grid.messages.GetAssociationRequestResponse.ContentEntry\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0e\n\x06source\x18\x04 \x01(\t\x12\x0e\n\x06target\x18\x05 \x01(\t\x1a.\n\x0c\x43ontentEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\x97\x01\n\x1dGetAssociationRequestsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xf6\x02\n\x1eGetAssociationRequestsResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12V\n\x07\x63ontent\x18\x02 \x03(\x0b\x32\x45.syft.grid.messages.GetAssociationRequestsResponse.metadata_container\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address\x1a\xac\x01\n\x12metadata_container\x12\x65\n\x08metadata\x18\x01 \x03(\x0b\x32S.syft.grid.messages.GetAssociationRequestsResponse.metadata_container.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xb1\x01\n\x1f\x44\x65leteAssociationRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x16\n\x0e\x61ssociation_id\x18\x03 \x01(\x05\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)


_SENDASSOCIATIONREQUESTMESSAGE = DESCRIPTOR.message_types_by_name[
    "SendAssociationRequestMessage"
]
_SENDASSOCIATIONREQUESTMESSAGE_METADATAENTRY = (
    _SENDASSOCIATIONREQUESTMESSAGE.nested_types_by_name["MetadataEntry"]
)
_RECEIVEASSOCIATIONREQUESTMESSAGE = DESCRIPTOR.message_types_by_name[
    "ReceiveAssociationRequestMessage"
]
_RECEIVEASSOCIATIONREQUESTMESSAGE_METADATAENTRY = (
    _RECEIVEASSOCIATIONREQUESTMESSAGE.nested_types_by_name["MetadataEntry"]
)
_RESPONDASSOCIATIONREQUESTMESSAGE = DESCRIPTOR.message_types_by_name[
    "RespondAssociationRequestMessage"
]
_GETASSOCIATIONREQUESTMESSAGE = DESCRIPTOR.message_types_by_name[
    "GetAssociationRequestMessage"
]
_GETASSOCIATIONREQUESTRESPONSE = DESCRIPTOR.message_types_by_name[
    "GetAssociationRequestResponse"
]
_GETASSOCIATIONREQUESTRESPONSE_CONTENTENTRY = (
    _GETASSOCIATIONREQUESTRESPONSE.nested_types_by_name["ContentEntry"]
)
_GETASSOCIATIONREQUESTSMESSAGE = DESCRIPTOR.message_types_by_name[
    "GetAssociationRequestsMessage"
]
_GETASSOCIATIONREQUESTSRESPONSE = DESCRIPTOR.message_types_by_name[
    "GetAssociationRequestsResponse"
]
_GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER = (
    _GETASSOCIATIONREQUESTSRESPONSE.nested_types_by_name["metadata_container"]
)
_GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER_METADATAENTRY = (
    _GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER.nested_types_by_name[
        "MetadataEntry"
    ]
)
_DELETEASSOCIATIONREQUESTMESSAGE = DESCRIPTOR.message_types_by_name[
    "DeleteAssociationRequestMessage"
]
SendAssociationRequestMessage = _reflection.GeneratedProtocolMessageType(
    "SendAssociationRequestMessage",
    (_message.Message,),
    {
        "MetadataEntry": _reflection.GeneratedProtocolMessageType(
            "MetadataEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _SENDASSOCIATIONREQUESTMESSAGE_METADATAENTRY,
                "__module__": "proto.grid.messages.association_messages_pb2"
                # @@protoc_insertion_point(class_scope:syft.grid.messages.SendAssociationRequestMessage.MetadataEntry)
            },
        ),
        "DESCRIPTOR": _SENDASSOCIATIONREQUESTMESSAGE,
        "__module__": "proto.grid.messages.association_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.SendAssociationRequestMessage)
    },
)
_sym_db.RegisterMessage(SendAssociationRequestMessage)
_sym_db.RegisterMessage(SendAssociationRequestMessage.MetadataEntry)

ReceiveAssociationRequestMessage = _reflection.GeneratedProtocolMessageType(
    "ReceiveAssociationRequestMessage",
    (_message.Message,),
    {
        "MetadataEntry": _reflection.GeneratedProtocolMessageType(
            "MetadataEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _RECEIVEASSOCIATIONREQUESTMESSAGE_METADATAENTRY,
                "__module__": "proto.grid.messages.association_messages_pb2"
                # @@protoc_insertion_point(class_scope:syft.grid.messages.ReceiveAssociationRequestMessage.MetadataEntry)
            },
        ),
        "DESCRIPTOR": _RECEIVEASSOCIATIONREQUESTMESSAGE,
        "__module__": "proto.grid.messages.association_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.ReceiveAssociationRequestMessage)
    },
)
_sym_db.RegisterMessage(ReceiveAssociationRequestMessage)
_sym_db.RegisterMessage(ReceiveAssociationRequestMessage.MetadataEntry)

RespondAssociationRequestMessage = _reflection.GeneratedProtocolMessageType(
    "RespondAssociationRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _RESPONDASSOCIATIONREQUESTMESSAGE,
        "__module__": "proto.grid.messages.association_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.RespondAssociationRequestMessage)
    },
)
_sym_db.RegisterMessage(RespondAssociationRequestMessage)

GetAssociationRequestMessage = _reflection.GeneratedProtocolMessageType(
    "GetAssociationRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETASSOCIATIONREQUESTMESSAGE,
        "__module__": "proto.grid.messages.association_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetAssociationRequestMessage)
    },
)
_sym_db.RegisterMessage(GetAssociationRequestMessage)

GetAssociationRequestResponse = _reflection.GeneratedProtocolMessageType(
    "GetAssociationRequestResponse",
    (_message.Message,),
    {
        "ContentEntry": _reflection.GeneratedProtocolMessageType(
            "ContentEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _GETASSOCIATIONREQUESTRESPONSE_CONTENTENTRY,
                "__module__": "proto.grid.messages.association_messages_pb2"
                # @@protoc_insertion_point(class_scope:syft.grid.messages.GetAssociationRequestResponse.ContentEntry)
            },
        ),
        "DESCRIPTOR": _GETASSOCIATIONREQUESTRESPONSE,
        "__module__": "proto.grid.messages.association_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetAssociationRequestResponse)
    },
)
_sym_db.RegisterMessage(GetAssociationRequestResponse)
_sym_db.RegisterMessage(GetAssociationRequestResponse.ContentEntry)

GetAssociationRequestsMessage = _reflection.GeneratedProtocolMessageType(
    "GetAssociationRequestsMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETASSOCIATIONREQUESTSMESSAGE,
        "__module__": "proto.grid.messages.association_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetAssociationRequestsMessage)
    },
)
_sym_db.RegisterMessage(GetAssociationRequestsMessage)

GetAssociationRequestsResponse = _reflection.GeneratedProtocolMessageType(
    "GetAssociationRequestsResponse",
    (_message.Message,),
    {
        "metadata_container": _reflection.GeneratedProtocolMessageType(
            "metadata_container",
            (_message.Message,),
            {
                "MetadataEntry": _reflection.GeneratedProtocolMessageType(
                    "MetadataEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER_METADATAENTRY,
                        "__module__": "proto.grid.messages.association_messages_pb2"
                        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetAssociationRequestsResponse.metadata_container.MetadataEntry)
                    },
                ),
                "DESCRIPTOR": _GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER,
                "__module__": "proto.grid.messages.association_messages_pb2"
                # @@protoc_insertion_point(class_scope:syft.grid.messages.GetAssociationRequestsResponse.metadata_container)
            },
        ),
        "DESCRIPTOR": _GETASSOCIATIONREQUESTSRESPONSE,
        "__module__": "proto.grid.messages.association_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetAssociationRequestsResponse)
    },
)
_sym_db.RegisterMessage(GetAssociationRequestsResponse)
_sym_db.RegisterMessage(GetAssociationRequestsResponse.metadata_container)
_sym_db.RegisterMessage(GetAssociationRequestsResponse.metadata_container.MetadataEntry)

DeleteAssociationRequestMessage = _reflection.GeneratedProtocolMessageType(
    "DeleteAssociationRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEASSOCIATIONREQUESTMESSAGE,
        "__module__": "proto.grid.messages.association_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.DeleteAssociationRequestMessage)
    },
)
_sym_db.RegisterMessage(DeleteAssociationRequestMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SENDASSOCIATIONREQUESTMESSAGE_METADATAENTRY._options = None
    _SENDASSOCIATIONREQUESTMESSAGE_METADATAENTRY._serialized_options = b"8\001"
    _RECEIVEASSOCIATIONREQUESTMESSAGE_METADATAENTRY._options = None
    _RECEIVEASSOCIATIONREQUESTMESSAGE_METADATAENTRY._serialized_options = b"8\001"
    _GETASSOCIATIONREQUESTRESPONSE_CONTENTENTRY._options = None
    _GETASSOCIATIONREQUESTRESPONSE_CONTENTENTRY._serialized_options = b"8\001"
    _GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER_METADATAENTRY._options = None
    _GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER_METADATAENTRY._serialized_options = (
        b"8\001"
    )
    _SENDASSOCIATIONREQUESTMESSAGE._serialized_start = 139
    _SENDASSOCIATIONREQUESTMESSAGE._serialized_end = 454
    _SENDASSOCIATIONREQUESTMESSAGE_METADATAENTRY._serialized_start = 407
    _SENDASSOCIATIONREQUESTMESSAGE_METADATAENTRY._serialized_end = 454
    _RECEIVEASSOCIATIONREQUESTMESSAGE._serialized_start = 457
    _RECEIVEASSOCIATIONREQUESTMESSAGE._serialized_end = 796
    _RECEIVEASSOCIATIONREQUESTMESSAGE_METADATAENTRY._serialized_start = 407
    _RECEIVEASSOCIATIONREQUESTMESSAGE_METADATAENTRY._serialized_end = 454
    _RESPONDASSOCIATIONREQUESTMESSAGE._serialized_start = 799
    _RESPONDASSOCIATIONREQUESTMESSAGE._serialized_end = 1003
    _GETASSOCIATIONREQUESTMESSAGE._serialized_start = 1006
    _GETASSOCIATIONREQUESTMESSAGE._serialized_end = 1180
    _GETASSOCIATIONREQUESTRESPONSE._serialized_start = 1183
    _GETASSOCIATIONREQUESTRESPONSE._serialized_end = 1454
    _GETASSOCIATIONREQUESTRESPONSE_CONTENTENTRY._serialized_start = 1408
    _GETASSOCIATIONREQUESTRESPONSE_CONTENTENTRY._serialized_end = 1454
    _GETASSOCIATIONREQUESTSMESSAGE._serialized_start = 1457
    _GETASSOCIATIONREQUESTSMESSAGE._serialized_end = 1608
    _GETASSOCIATIONREQUESTSRESPONSE._serialized_start = 1611
    _GETASSOCIATIONREQUESTSRESPONSE._serialized_end = 1985
    _GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER._serialized_start = 1813
    _GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER._serialized_end = 1985
    _GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER_METADATAENTRY._serialized_start = (
        407
    )
    _GETASSOCIATIONREQUESTSRESPONSE_METADATA_CONTAINER_METADATAENTRY._serialized_end = (
        454
    )
    _DELETEASSOCIATIONREQUESTMESSAGE._serialized_start = 1988
    _DELETEASSOCIATIONREQUESTMESSAGE._serialized_end = 2165
# @@protoc_insertion_point(module_scope)
